"""
File: explain_kan.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Explain KAN models by calculating edge relevance scores in the same way as Ziming Liu's pyKAN 2.0
"""
import torch

from ..models.wrapper import WrapperTemplate


class KANExplainer():
    """
    Calculate relevances scores for given model using the same approach as in Ziming Liu's pyKAN 2.0.
    """

    def __init__(self, model: WrapperTemplate, samples, method="pykan"):
        """
        :param model: Model to explain
        :param samples: samples to explain the model on
        :param method: method for explaining (currently only pykan is supported)
        """
        self.model = model
        self.method = method
        # create dictionary to store edge and node relevance scores
        self.node_relevances = dict()
        self.edge_relevances = dict()
        # remember samples
        self.samples = samples
        if method == "pykan":
            self.calc_relevances_pykan()
        else:
            raise NotImplementedError("Method {} not implemented".format(method))

    def calc_relevances_pykan(self):
        """
        Calculate the relevance scores of the model by using the edge and node acitvations stored within the model
        (WrapperTemplate)
        """
        # collect activations of model's edges and nodes for the samples specified
        with torch.no_grad():
            self.model.gather_activations(self.samples.to("cuda"))
        num_layers = len(self.model.get_kan_layers())
        # iterate through the layers in reverse
        for layer_idx in reversed(range(num_layers)):
            layer = self.model.get_kan_layers()[layer_idx]
            num_neurons_in_layer = self.model.get_layersizes()[layer_idx]
            num_neurons_out_layer = self.model.get_layersizes()[layer_idx + 1]
            if layer_idx == num_layers - 1:  # last layer
                # Set Output Node Relevances to 1
                for j in range(num_neurons_out_layer):  # possibly multiple output neurons
                    self.node_relevances[(layer_idx+1, j)] = 1
            # iterate over all neurons in the previous layer
            for i in range(num_neurons_in_layer):
                outgoing_relevances = 0  # outgoing relevance of the current node i in layer (layer_idx -1)
                for j in range(num_neurons_out_layer):  # look at edge E_{layer_idx-1, i, j}
                    # Formula 9 from Kan 2.0 Paper with (probably) fixed Notation
                    a_lpj = self.node_relevances[(layer_idx + 1, j)]
                    # calculate edge and node score as they did in KAN 2.0 paper
                    e_lm_ij = torch.std(self.model.get_edge_activations((layer_idx, i, j)), dim=0)
                    n_lp_j = torch.std(self.model.get_node_activations((layer_idx + 1, j)), dim=0)
                    b_lij =  a_lpj * e_lm_ij / n_lp_j
                    # sum up all outgoing edge's relevance scores for source node's relevance score
                    outgoing_relevances += b_lij
                    # store edge relevance b_lij
                    self.edge_relevances[(layer_idx, i, j)] = b_lij
                self.node_relevances[(layer_idx, i)] = outgoing_relevances
    def get_edge_relevance(self, l_i_j_index):
        """
        Returns edge relevance score for edge connection node i in layer l with node j in the next layer
        :param l_i_j_index: tuple (l,i,j)
        :return: relevance score of edge E_{l,i,j}
        """
        return self.edge_relevances[l_i_j_index]
    def get_node_relevance(self, l_i_index):
        """
        Returns node relevance score for node i in layer l
        :param l_i_index: tuple (l, i)
        :return: relevance score of node N_{l,i}
        """
        return self.node_relevances[l_i_index]


class DummyKANExplainer(KANExplainer):
    """
    Dummy KAN Explainer that always returns 1 (placeholder if gather_activations(), ... are not implemented by the model
    """
    def __init__(self):
        pass
    def calc_relevances_pykan(self):
        pass
    def get_edge_relevance(self, l_i_j_index):
        return 1
    def get_node_relevance(self, l_i_index):
        return 1


"""
File: CVKANWrapper.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Wrapper for CVKAN to make it compatible with our KanPlotter and KanExplainer.
"""
import numpy as np
import torch

from ..CVKAN import CVKAN
from ..functions.ComplexSilu import complex_silu_realweights, complex_silu_complexweight
from .WrapperTemplate import WrapperTemplate


class CVKANWrapper(CVKAN, WrapperTemplate):
    def __init__(self, **kwargs):
        # forward all arguments to superclass (CVKAN)
        super().__init__(**kwargs)
        self.edge_activations = dict()
        self.node_activations = dict()
    def plot_curve(self, l_i_j_index: tuple, num_pts=100) -> (np.ndarray, tuple[np.ndarray]):
        with torch.no_grad():
            l, i, j = list(l_i_j_index)
            # get grid ranges
            grid_min, grid_max = self.get_gridrange(l)
            # construct mesh grid to cover the whole grid evenly
            reals, imags = np.meshgrid(np.linspace(grid_min, grid_max, num_pts),
                                       np.linspace(grid_min, grid_max, num_pts))
            # turn meshgrid into complex numbers
            inputs = torch.complex(torch.from_numpy(reals), torch.from_numpy(imags)).to(torch.complex64)
            # move them to correct device
            inputs = inputs.flatten().to(self.get_kan_layers()[0].realweights.device)

            # grid has shape num_grids x num_grids
            # transform x to shape BATCH x Input-Dim x num_grids x num_grids
            x = inputs.unsqueeze(-1).unsqueeze(-1).expand(inputs.shape + (self.get_kan_layers()[l].num_grids, self.get_kan_layers()[l].num_grids))
            # apply RBF on x (centered around each grid point)
            result = torch.exp(-(torch.abs(x - self.get_kan_layers()[l].grid)) ** 2 / self.get_kan_layers()[l].rho)
            # u and v are the two grid dimensions
            result_real = torch.einsum("buv,uv->b", result, self.get_kan_layers()[l].realweights[i, j, :, :])
            result_imag = torch.einsum("buv,uv->b", result, self.get_kan_layers()[l].complexweights[i, j, :, :])
            # construct complex number out of real and imaginary parts again
            outputs = torch.complex(result_real, result_imag)
            # kind of CSiLU used
            if self.csilu_type == "complex_weight":
                outputs = outputs + self.get_kan_layers()[l].silu_weight[i,j] * complex_silu_complexweight(inputs)
            elif self.csilu_type == "real_weights":
                outputs = outputs + torch.complex(self.get_kan_layers()[l].silu_weight[i, j, 0] * complex_silu_realweights(inputs)[0],
                                                  self.get_kan_layers()[l].silu_weight[i, j, 1] * complex_silu_realweights(inputs)[1])
            # Add SiLU Bias
            outputs +=  self.get_kan_layers()[l].silu_bias[i,j]
            outputs = torch.reshape(outputs, (num_pts, num_pts))
            # TODO include basis and RBFs separately (who cares? maybe for visualization purposes...)
            basis = None
            rbfs = None
        return (reals, imags), (basis, rbfs, outputs.detach().cpu().numpy())

    def get_edge_activations(self, l_i_j):
        l, i, j = l_i_j
        return self.edge_activations[l][:, i, j]

    def get_node_activations(self, l_i):
        l, i = l_i
        return self.node_activations[l][:, i]

    def get_layersizes(self) -> list:
        return self.layers_hidden

    def get_gridrange(self, layer_index) -> list:
        layer = self.layers[layer_index]
        return [layer.grid_min, layer.grid_max]

    def get_kan_layers(self) -> torch.nn.ModuleList:
        return self.layers

    def gather_activations(self, x):
        for layer_idx, layer in enumerate(self.get_kan_layers()):
            # current layer's input are Node activations
            self.node_activations[layer_idx] = x.detach().cpu()
            # x has shape BATCH x Input-Dim
            assert len(x.shape) == 2 and x.shape[1] == layer.input_dim
            # grid has shape num_grids x num_grids
            # transform x to shape BATCH x Input-Dim x num_grids x num_grids
            x_expanded = x.unsqueeze(-1).unsqueeze(-1).expand(x.shape + (layer.num_grids, layer.num_grids,))
            # apply RBF on x_expanded (centered around each grid point)
            result = torch.exp(-(torch.abs(x_expanded - layer.grid)) ** 2 / layer.rho)
            # u and v are the two grid dimensions
            # i and o are input and output indices within layer layer_idx and layer_ix+1
            result_real = torch.einsum("biuv,iouv->bio", result, layer.realweights)
            result_imag = torch.einsum("biuv,iouv->bio", result, layer.complexweights)
            # construct complex number out of real and imaginary parts again
            result_complex = torch.complex(result_real, result_imag)
            assert result_complex.shape[2] == layer.output_dim

            # kind of CSiLU used
            if layer.csilu_type == "complex_weight":
                silu_value = torch.einsum("io,bi->bio", layer.silu_weight, complex_silu_complexweight(x))
            elif layer.csilu_type == "real_weights":
                silu_value_raw = complex_silu_realweights(x)
                silu_value = torch.complex(torch.einsum("io,bi->bio", layer.silu_weight[:, :, 0], silu_value_raw[0]),
                                           torch.einsum("io,bi->bio", layer.silu_weight[:, :, 1], silu_value_raw[1]))
            # Add complex CSiLU bias
            silu_value += layer.silu_bias
            result_complex += silu_value
            # store Edge activations
            self.edge_activations[layer_idx] = result_complex.detach().cpu()
            # sum over all incoming Edges into each Node
            result_complex = torch.einsum("bio->bo", result_complex)
            # potentially apply Normalization
            if layer.use_norm is not None:
                result_complex = layer.norm(result_complex)

            x = result_complex
        self.node_activations[len(self.layers)] = x  # output activations


"""
File: cplot.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Experiments with plotting standard complex-valued functions (i.e. z^2)
"""
import matplotlib
import cplotting_tools as cplt
import numpy as np

matplotlib.use('TkAgg')  # use a Backend that supports interactive elements

# create meshgrid for range [-2, 2] x [-2, 2]
xs, ys = np.meshgrid(np.linspace(-2,2,100), np.linspace(-2,2,100))

# convert meshgrid to complex-valued numbers (real and imaginary parts)
zs = xs + 1j*ys

# calculate result
f = zs**2
#f = np.sin(zs)
#f = np.exp(3*zs)

# plot result
cplt.complex_plot3D(xs, ys, f)
cplt.plot_re_im(xs, ys, f)
#cplt.domain_coloring(xs,ys,f,cmap="hsv")



"""
File: CVKAN.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: CVKAN Model definition
"""
from typing import List
import torch


from .functions.CV_LayerNorm import Complex_LayerNorm, Complex_BatchNorm, Complex_BatchNorm_naiv, \
    Complex_BatchNorm_var

from .functions.ComplexSilu import complex_silu_complexweight, complex_silu_realweights

class Norms:
    """Enum for Normalization Types"""
    LayerNorm = "layernorm"
    BatchNorm = "batchnorm"  # BN_{\mathbb{C}}
    BatchNormNaiv = "batchnormnaiv"  # BN_{\mathbb{R}^2}
    BatchNormVar = "batchnormvar"  # BN_{\mathbb{V}} using variance
    NoNorm = None


class CVKANLayer(torch.nn.Module):
    def __init__(self, input_dim: int, output_dim: int, num_grids: int = 8, grid_min = -2, grid_max = 2, rho=1, use_norm=Norms.BatchNorm, csilu_type="complex_weight"):
        """
        :param input_dim: input dimension size of Layer (Layer Width)
        :param output_dim: output dimension size of Layer (next Layer's Width)
        :param num_grids: number of grid points ***per dimension***; thus num_grids * num_grids gridpoints in total
        :param grid_min: left limit of grid
        :param grid_max: right limit of grid
        :param rho: rho for use in RBF (default rho=1)
        :param use_norm: which Normalization scheme to use
        :param csilu_type: the kind of CSiLU to use ('complex_weight' or 'real_weights')
        """
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_grids = num_grids
        self.grid_min = grid_min
        self.grid_max = grid_max
        self.csilu_type = csilu_type
        self.use_norm = use_norm
        # initialize Norm instance corresponding to self.use_norm
        if self.use_norm == Norms.LayerNorm:
            self.norm = Complex_LayerNorm()
        elif self.use_norm == Norms.BatchNorm:
            self.norm = Complex_BatchNorm(num_channel=output_dim)
        elif self.use_norm == Norms.BatchNormNaiv:
            self.norm = Complex_BatchNorm_naiv(num_channel=output_dim)
        elif self.use_norm == Norms.BatchNormVar:
            self.norm = Complex_BatchNorm_var(num_channel=output_dim)
        elif self.use_norm == Norms.NoNorm:
            self.norm = None
        else:
            raise NotImplementedError()
        # create grid points 2D array
        real = torch.linspace(grid_min, grid_max, num_grids)
        real = real.unsqueeze(1).expand(num_grids, num_grids)

        imag = torch.linspace(grid_min, grid_max, num_grids)
        imag = imag.unsqueeze(0).expand(num_grids, num_grids)
        # make it complex-valued from real and imaginary parts
        grid = torch.complex(real, imag)
        # grid is a non-trainable Parameter
        self.grid = torch.nn.Parameter(grid, requires_grad=False)
        self.rho = rho
        # weights for each RBF centered around the grid points
        self.realweights = torch.nn.Parameter(torch.randn(size=(input_dim, output_dim, num_grids, num_grids)), requires_grad=True)
        self.complexweights = torch.nn.Parameter(torch.randn(size=(input_dim, output_dim, num_grids, num_grids)), requires_grad=True)
        # initialize CSiLU weight to use based on selected csilu_type
        if self.csilu_type == "complex_weight":
            self.silu_weight = torch.nn.Parameter(torch.ones(size=(self.input_dim, self.output_dim), dtype=torch.complex64), requires_grad=True)
        elif self.csilu_type == "real_weights":
            self.silu_weight = torch.nn.Parameter(torch.ones(size=(self.input_dim, self.output_dim, 2), dtype=torch.float32), requires_grad=True)
        else:
            raise NotImplementedError()
        # add complex-valued bias to CSiLU
        self.silu_bias = torch.nn.Parameter(torch.zeros(size=(input_dim, output_dim), dtype=torch.complex64), requires_grad=True)


    def forward(self, x):
        # x has shape BATCH x Input-Dim
        assert len(x.shape) == 2 and x.shape[1] == self.input_dim, f"Wrong Input Dimension! Got {x.shape} for Layer with Dimensions[{self.input_dim}, {self.output_dim}]"
        # grid has shape num_grids x num_grids
        # transform x to shape BATCH x Input-Dim x num_grids x num_grids
        x_expanded = x.unsqueeze(-1).unsqueeze(-1).expand(x.shape + (self.num_grids, self.num_grids,))
        # apply RBF on x (centered around each grid point)
        result = torch.exp(-(torch.abs(x_expanded - self.grid)) ** 2 / self.rho)
        # u and v are the two grid dimensions
        # i and o are input and output indices within layer layer_idx and layer_ix+1
        result_real = torch.einsum("biuv,iouv->bo", result, self.realweights)
        result_imag = torch.einsum("biuv,iouv->bo", result, self.complexweights)
        # construct complex number out of real and imaginary parts again
        result_complex = torch.complex(result_real, result_imag)
        assert result_complex.shape[1] == self.output_dim, f"Wrong Output Dimension! Got {result_complex.shape} for Layer with Dimensions[{self.input_dim}, {self.output_dim}]"
        # kind of CSiLU used
        if self.csilu_type == "complex_weight":
            silu_value = torch.einsum("io,bi->bio",self.silu_weight, complex_silu_complexweight(x))
        elif self.csilu_type == "real_weights":
            silu_value_raw = complex_silu_realweights(x)
            silu_value = torch.complex(torch.einsum("io,bi->bio",self.silu_weight[:,:,0], silu_value_raw[0]),
                                       torch.einsum("io,bi->bio",self.silu_weight[:,:,1], silu_value_raw[1]))
        # Add complex CSiLU bias
        silu_value += self.silu_bias
        silu_value = torch.einsum("bio->bo", silu_value)
        result_complex = result_complex + silu_value
        # potentially apply Normalization
        if self.use_norm is not None:
            result_complex = self.norm(result_complex)
        return result_complex

    def to(self, device):
        super().to(device)
        # move grid to the right device
        self.grid = self.grid.to(device)
        if self.use_norm is not None:  # and Norm as well
            self.norm = self.norm.to(device)
class CVKAN(torch.nn.Module):
    def __init__(self,
                 layers_hidden: List[int],
                 num_grids: int = 8,
                 rho=1,
                 use_norm=Norms.BatchNorm,
                 grid_mins = -2,
                 grid_maxs = 2,
                 csilu_type = "complex_weight"):
        """
        :param layers_hidden: List with Layer Sizes (i.e. [1,5,3,1] for a 1x5x3x1 CVKAN)
        :param num_grids: Number of Grid Points ***per dimension*** (so in total num_grids * num_grids gridpoints)
        :param rho: rho for RBF (default rho=1)
        :param use_norm: which Normalization scheme to use. Normalization is applied AFTER every layer except the last
        :param grid_mins: left limit of grid
        :param grid_maxs: right limit of grid
        :param csilu_type: type of CSiLU to use ('complex_weight' or 'real_weights')
        """
        super().__init__()
        # convert grid limits to list if not already is list (limits for each layer independently)
        if not type(grid_mins) == list:
            grid_mins = [grid_mins] * len(layers_hidden)
        if not type(grid_maxs) == list:
            grid_maxs = [grid_maxs] * len(layers_hidden)
        self.layers_hidden = layers_hidden
        self.num_grids = num_grids
        self.rho = rho
        self.use_norm = use_norm
        self.csilu_type = csilu_type
        # convert csilu_type to list (each layer could get it's own CSiLU type)
        if type(self.use_norm) != list:
            self.use_norm = [self.use_norm] * (len(layers_hidden) - 1)
        else:
            assert len(self.use_norm) == len(self.layers_hidden)
        # lambdas to calculate if Layer i should have Normalization applied after it
        is_last_layer = lambda i: i >= len(self.layers_hidden) - 2
        norm_to_use = lambda i: self.use_norm[i] if not is_last_layer(i) else Norms.NoNorm
        # Array with Normalization schemes to use after every layer
        self.use_norm = [norm_to_use(i) for i in range(len(layers_hidden)-1)]
        # stack Layers into a ModuleList
        self.layers = torch.nn.ModuleList([CVKANLayer(input_dim=layers_hidden[i], output_dim=layers_hidden[i+1],
                                                      num_grids=num_grids, grid_min=grid_mins[i], grid_max=grid_maxs[i],
                                                      rho=self.rho, use_norm=norm_to_use(i),
                                                      csilu_type=self.csilu_type) for i in range(len(layers_hidden) - 1)])
    def forward(self, x):
        # make sure x is batched
        if len(x.shape) == 1:
            x = x.unsqueeze(1)
        # find mins over all samples within batch
        mins_per_channel = torch.minimum(torch.amin(x.real, dim=0), torch.amin(x.imag, dim=0))
        maxs_per_channel = torch.maximum(torch.amax(x.real, dim=0), torch.amax(x.imag, dim=0))
        # make sure first Layer's grid limits aren't overstepped. This would indicate a problem with dataset
        # normalization (which must be done before entering the data into the model!)
        assert (mins_per_channel >= self.layers[0].grid_min).all() and (maxs_per_channel <= self.layers[0].grid_max).all(), "Input data does not fall completely within the grid range of the first layer. Please normalize the data!"
        # feed data through the layers
        for layer in self.layers:
            x = layer(x)
        return x
    def to(self, device):
        for layer in self.layers:
            layer.to(device)



"""
File: cplotting_tools.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang; original author: Arturo Mena López
Description: Modified copy of cplotting_tools.py by Arturo Mena López
             (see https://github.com/artmenlope/complex-plotting-tools/blob/master/cplotting_tools.py ; 17.01.2025)
             Changes include (limited to method 'complex_plot3D', everything else remains unchanged):
                - Added support to draw directly into figure given as parameter instead of creating new figure
                - added support to return an image of the plot
                - made fontsize a parameter
                - introduced a flag whether colormap should be drawn
                - added support to return all drawn matplotlib.pyplot LineSet objects (to later toggle their
                  transparency in KanPlot()

"""
"""
Created on Tue Jul  7 13:05:40 2020

@author: artmenlope
"""

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import cmath



def colorize(f, a=0.5, log_brightness=True, log_contrast=0.4):
    
    """
    Auxiliar function for creating domain coloring plots.
    
    Given the evaluated function f, returns an array of colors
    representing the function's phase. 
    
    The resulting colors encode the module of the function as 
    the brightness.

    Arguments:

        f :: 2D numpy array of complex numbers. Evaluated 
             function to be plotted.

        a :: Float between 0 and 1. Parameter for the brightness.

        log_brightness :: Boolean. If True, the module of f is
                          represented via the brightness of the
                          colors in a logarithmic way. If False,
                          The brightness changes exponentially as
                          the module of f increases.

        log_contrast :: Float. Parameter for the brightness.
    """

    from colorsys import hls_to_rgb
    
    def logb(arg, base):
        """Return the logarithm with base b of arg."""
        return np.log(base) / np.log(base)
       
    H = (np.pi-np.arctan2(f.imag, -f.real))/(2*np.pi) # Hue.
   
    if log_brightness == False:
        L = (1-a**np.abs(f)) # Brightness.
        
    if log_brightness == True:
        L = 1-a**np.log(1+np.abs(f)**log_contrast)
        
    S = 1 # Saturation.
    
    c = np.vectorize(hls_to_rgb)(H, L, S) # --> Tuple.
    c = np.array(c)  # The array of colors c is of shape (3,n,m), but it needs to be (m,n,3).
    c = np.rot90(c.transpose(2,1,0), 1) # Change shape to (m,n,3) and rotate 90 degrees as correction.
    
    return c



def domain_coloring(x, y, f, 
                   figsize=(12,8),
                   xlabel="Re", 
                   ylabel="Im",
                   title=None,
                   grid=False,
                   cmap="hsv"):
    
    """
    Domain coloring plot. 
    
    The evaluated function f is transformed to polar form and its 
    phase is represented using colors. The module is not represented.
    
    See https://en.wikipedia.org/wiki/Domain_coloring for more 
    information.

    x, y, f are 2D arrays. f can contain complex numbers.
    figsize, xlabel, ylabel, title, grid and cmap are parameters 
    for the Matplotlib plot.
    """
    
    arg_f = np.mod(np.angle(f),2*np.pi) # np.mod ensures argument from 0 to 2*pi
    lim = np.max([x, y])

    # Prepare for using colormaps.
    norm = matplotlib.colors.Normalize(vmin=0,vmax=2*np.pi)
    c_m = cmap # "twilight", "hsv"
    s_m = matplotlib.cm.ScalarMappable(cmap=c_m, norm=norm)
    s_m.set_array([])
    fcolors = s_m.to_rgba(arg_f)
    
    # A figure and a 3d subplot.
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.set_xlabel(xlabel, fontsize=14)
    ax.set_ylabel(ylabel, fontsize=14)
    
    # Limit corrections.
    ax.set_xlim((-lim,lim))
    ax.set_ylim((-lim,lim))
    
    # Grid and title.
    ax.grid(grid)
    
    if title is not None:
        ax.set_title(title, fontsize=18, pad=20, usetex=False)
        
    ax.imshow(arg_f, cmap=cmap, extent=[-lim,lim,-lim,lim], interpolation="none", origin="lower")
   
    # Draw the colorbar.
    cbar = plt.colorbar(s_m, ticks=[0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], pad=0.1)
    cbar.ax.set_yticklabels(["$0$", "$\\frac{\\pi}{2}$", "$\\pi$", "$\\frac{3\\pi}{2}$", "$2\\pi$"], fontsize=16)
    
    plt.tight_layout()
    plt.show()



def domain_coloring_illuminated(x, y, f, 
                                a = 0.5,
                                log_brightness=True,
                                log_contrast=0.4,
                                figsize=(12,8),
                                xlabel="Re", 
                                ylabel="Im",
                                title=None,
                                grid=False):
    
    """
    Domain coloring plot. 
    
    The function f is transformed to polar form and its phase 
    is represented using colors. The module is represented using
    the brightness of the colors.
    
    See https://en.wikipedia.org/wiki/Domain_coloring for more 
    information.

    Arguments:

        x, y :: 2D arrays. They represent the 2D plotting space.

        f :: 2D numpy array of complex numbers. Evaluated 
             function to be plotted.

        a :: Float between 0 and 1. Parameter for the brightness.

        log_brightness :: Boolean. If True, the module of f is
                          represented via the brightness of the
                          colors in a logarithmic way. If False,
                          The brightness changes exponentially as
                          the module of f increases.

        log_contrast :: Float. Parameter for the brightness.

        figsize, xlabel, ylabel, title and grid are parameters 
        for the Matplotlib plot.
    """
    
    img = colorize(f, a, log_brightness, log_contrast)
    arg_f = np.mod(np.angle(f),2*np.pi) # np.mod ensures argument from 0 to 2*pi
    lim = np.max([x, y])

    # initializing the colormap machinery
    norm = matplotlib.colors.Normalize(vmin=0,vmax=2*np.pi)
    c_m = "hsv"
    s_m = matplotlib.cm.ScalarMappable(cmap=c_m, norm=norm)
    s_m.set_array([])
    
    # a figure and a 3d subplot
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.set_xlabel(xlabel, fontsize=14)
    ax.set_ylabel(ylabel, fontsize=14)
    
    # Limit corrections
    ax.set_xlim((-lim,lim))
    ax.set_ylim((-lim,lim))
    
    # Grid and title
    ax.grid(grid)
    
    if title is not None:
        ax.set_title(title, fontsize=18, pad=20, usetex=False)
    
    #ax.contourf(x, y, arg_f, cmap="hsv", levels=50, alpha=1)
    ax.imshow(img, extent=[-lim,lim,-lim,lim], interpolation="none", origin="upper")
   
    # Draw the colorbar 
    cbar = plt.colorbar(s_m, ticks=[0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], pad=0.1)
    cbar.ax.set_yticklabels(["$0$", "$\\frac{\\pi}{2}$", "$\\pi$", "$\\frac{3\\pi}{2}$", "$2\\pi$"], fontsize=16)
    
    plt.tight_layout()
    plt.show()
    
    

def complex_plot3D(x, y, f, 
                   figsize=(12,8),
                   f_lim=10,
                   offset=0,
                   xlabel="Re", 
                   ylabel="Im", 
                   zlabel="$|f(z)|$",
                   title=None,
                   grid=True,                    
                   contour3D=False,
                   log_mode=True,
                   fig_ax = None,
                   draw_colorbar=True,
                   fontsize=14,
                   return_image=False):
    
    """
    3D plot representing he evaluated complex function f. 
    
    f is transformed to polar form. The module is represented 
    as a 3d surface and the phase is represented as colors over
    said surface like in a domain coloring plot. In addition, 
    the module of f is represented like shadows at the bottom
    of the plot. A darker shadow indicates lower values for
    the module.

    Arguments:

        x, y :: 2D arrays. They represent the 2D plotting space.

        f :: 2D numpy array of complex numbers. Evaluated 
             function to be plotted.

        f_lim :: Float greater than 0. Set the limit of the 
                 vertical axis. Improves the visualization in 
                 case the module diverges to infinity.

        contour3D :: Boolean. If True, plot a contour over the 
                     f module's plane at the bottom of the plot.

        log_mode :: Boolean. If True, the colors of the module's 
                    representation will increase in a logarithmic 
                    way.

        offset, xlabel, ylabel, zlabel, title and grid are parameters 
        for Matplotlib.
    """
    plotted_objects = []  # list to store LineSet objects (to later adjust transparency)
    if log_mode == True:
        abs_f = np.log2(np.abs(f)+1)
    if log_mode == False:
        abs_f = np.abs(f)
        
    arg_f = np.mod(np.angle(f),2*np.pi) # np.mod ensures argument from 0 to 2*pi
    lim = np.max([x, y])

    # initializing the colormap machinery
    norm = matplotlib.colors.Normalize(vmin=0,vmax=2*np.pi)
    c_m = matplotlib.cm.hsv #twilight, hsv
    s_m = matplotlib.cm.ScalarMappable(cmap=c_m, norm=norm)
    s_m.set_array([])
    fcolors = s_m.to_rgba(arg_f)
    
    if fig_ax is None:
        # create figure and 3d subplot
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection='3d')
    else:
        # use existing figure and axis given as parameter
        fig, ax = fig_ax
    # set axis labels
    ax.set_xlabel(xlabel, fontsize=fontsize, labelpad=fontsize//2)
    ax.set_ylabel(ylabel, fontsize=fontsize, labelpad=fontsize//2)
    ax.set_zlabel(zlabel, fontsize=fontsize, labelpad=fontsize)
    
    # Limit corrections
    abs_f[abs_f > f_lim] = f_lim
    lim = 0.96*lim
    ax.set_xlim((-lim,lim))
    ax.set_ylim((-lim,lim))
    ax.set_zlim((0,f_lim))
    
    # Grid and title
    ax.grid(grid)
    
    if title is not None:
        ax.set_title(title, fontsize=18, pad=20, usetex=False)
        
    # make the bottom pane transparent
    ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
    
    # Plot the modulus' surface with the argument as color.
    surf = ax.plot_surface(x, y, abs_f, linewidth=0, alpha=0.7,
                    cstride=1, rstride=1,
                    facecolors=fcolors)
    plotted_objects.append(surf)
    
    if contour3D == True:
        contour = ax.contour3D(x, y, abs_f, alpha=0.5, colors='black', levels=20)
        # contours do not support methods used on Artists. The Artists are in countour.collections attribute
        for artist in contour.allsegs:
            plotted_objects.append(artist)
    contour = ax.contourf(x, y, np.log2(abs_f+1), zdir='z', offset=offset  , cmap="gist_yarg_r", levels=50)
    # contours do not support methods used on Artists. The Artists are in countour.collections attribute
    for artist in contour.allsegs:
        plotted_objects.append(artist)
    
    ax.tick_params(axis='x', which='major', labelsize=fontsize, pad=fontsize//8)
    ax.tick_params(axis='y', which='major', labelsize=fontsize, pad=fontsize // 8)
    ax.tick_params(axis='z', which='major', labelsize=fontsize, pad=fontsize // 2)
   
    # Draw the colorbar
    if draw_colorbar:
        cbar = plt.colorbar(s_m, ticks=[0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], ax=ax)
        cbar.ax.set_yticklabels(["$0$", "$\\frac{\\pi}{2}$", "$\\pi$", "$\\frac{3\\pi}{2}$", "$2\\pi$"], fontsize=fontsize)
        cbar.ax.set_ylabel("Arg f(z)", fontsize=fontsize)
        plotted_objects.append(cbar)
    if fig_ax is None:
        if return_image:
            # save figure, read it using matplotlib and return it
            plt.savefig("temp.png", dpi=200, transparent=True)
            plt.close(fig)
            image= matplotlib.image.imread("temp.png")
            # create additional alpha channel for image (only neccessary for .jpg)
            #rgba_image = np.zeros((image.shape[0], image.shape[1], 4), dtype=image.dtype)
            # copy RGB without changes
            #rgba_image[:,:, :3] = image
            # set alpha channel to 255 if pixel is very white (> 0.99*255 intensity)
            #white_threshold = 0.9*255
            #mask = np.all(image > white_threshold, axis=-1)
            #rgba_image[..., 3] = np.where(mask, 1, 1)

            return image
        else:
            plt.show()
    else:  # if fig_ax is given: plot directly **into** the given figure and axis
        ax.yaxis.pane.set_visible(False)
        # append everything that's plotted to plotted_objects (se transparency can be adjusted later
        # for everything in this plot)
        plotted_objects.append(ax.xaxis.pane)
        plotted_objects.append(ax.yaxis.pane)
        plotted_objects.append(ax.zaxis.pane)

        plotted_objects.append(ax.xaxis.line)
        plotted_objects.append(ax.yaxis.line)
        plotted_objects.append(ax.zaxis.line)

        for axis in [ax.xaxis, ax.yaxis, ax.zaxis]:
            for ls in [axis.get_ticklabels(), axis.get_ticklines()]:
                for l in ls:
                    plotted_objects.append(l)
        plotted_objects.append(ax.xaxis.get_label())
        plotted_objects.append(ax.yaxis.get_label())
        plotted_objects.append(ax.zaxis.get_label())
        ax.grid(False)
        # data has been plotted into specified figure and axis
        # plotted lines will additionally be returned as a list (to later adjust transparency)
        return plotted_objects



def plot_re_im(x, y, f, 
               figsize=(14,7),
               alpha=1,
               # f_lims=None,
               title=None,
               grid=True,                    
               contour=False,
               cmap="viridis",
               synchronize_rotations=False):

    """
    Plot the real and the imaginary parts of the function 
    f in separated subplots as surfaces. The surfaces can 
    also be projected into a filled contour plot at the 
    bottom of the vertical axis.
    
    Arguments:

        x, y :: 2D arrays. They represent the 2D plotting 
                space.

        f :: 2D numpy array of complex numbers. Evaluated 
             function to be plotted.

        synchronize_rotations :: Boolean. If True, when using
                                 the Matplotlib's interactive
                                 plotting window and rotating
                                 a subplot, both subplots will 
                                 synchronize the rotation.

        figsize, alpha, title, grid, contour and cmap are 
        parameters for Matplotlib.
    """
    
    lim = np.max([x, y])
    
    # A figure and a 3d subplot
    fig = plt.figure(figsize=figsize)
    ax_re = fig.add_subplot(121, projection="3d")
    ax_im = fig.add_subplot(122, projection="3d")
    
    ax_re.set_xlabel("Re", fontsize=14)
    ax_re.set_ylabel("Im", fontsize=14)
    ax_re.set_title("Re $f(z)$", fontsize=18)
    
    ax_im.set_xlabel("Re", fontsize=14)
    ax_im.set_ylabel("Im", fontsize=14)
    ax_im.set_title("Im $f(z)$", fontsize=18)
    
    ax_re.set_xlim((-lim,lim))
    ax_re.set_ylim((-lim,lim))
    
    ax_im.set_xlim((-lim,lim))
    ax_im.set_ylim((-lim,lim))
    
    # Grid and title
    ax_re.grid(grid)
    ax_im.grid(grid)
    
    if title is not None:
        fig.suptitle(title, fontsize=18, usetex=False)
        
    # Plot function components.
    ax_re.plot_surface(x, y, f.real, linewidth=0, alpha=alpha,
                       cstride=1, rstride=1,
                       cmap=cmap)
    
    ax_im.plot_surface(x, y, f.imag, linewidth=0, alpha=alpha,
                       cstride=1, rstride=1,
                       cmap=cmap)
    
    if contour == True:
        ax_re.contourf(x, y, f.real, zdir='z', offset=ax_re.get_zlim()[0], cmap=cmap, levels=50, alpha=1)
        ax_im.contourf(x, y, f.imag, zdir='z', offset=ax_im.get_zlim()[0], cmap=cmap, levels=50, alpha=1)
    
    if synchronize_rotations == True:
        
        def on_move(event):
            if event.inaxes == ax_re:
                ax_im.view_init(elev=ax_re.elev, azim=ax_re.azim)
            elif event.inaxes == ax_im:
                ax_re.view_init(elev=ax_im.elev, azim=ax_im.azim)
            else:
                return
            fig.canvas.draw_idle()
    
        fig.canvas.mpl_connect('motion_notify_event', on_move)
        
    plt.tight_layout()
    plt.show()
    
    
    
def complex_vector_field(x, y, f,
                         figsize=(12,8),
                         title=None,
                         grid=False,
                         cmap=None,
                         dark_background=False,
                         norm=False):
    
    """
    Plot the complex function f as a 2D vector field.
    The plotted vectors have the form (Real(f), Imag(f)). 
    In polar form, the argument of the function f can
    be represented using a colormap in addition to the
    already visualized orientation of the vectors.

    Arguments:

        x, y :: 2D arrays. They represent the 2D plotting 
                space.

        f :: 2D numpy array of complex numbers. Evaluated 
             function to be plotted.

        dark_background :: Boolean. If True, sets the axis 
                           background color to black.

        norm :: Boolean. If True, normalizes the vectors.

        figsize, title, grid and cmap are parameters for 
        Matplotlib.
    """

    # Vector normalization.
    if norm == True:
        r = np.sqrt(f.real**2+f.imag**2)
        f = f.real/r + 1j*f.imag/r
    
    # Create the figure and axis.
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.set_aspect("equal")
    
    # Colormap.
    if cmap is None:
        
        ax.quiver(x, y, np.real(f), np.imag(f),  
                  color='blue', 
                  pivot="middle", 
                  norm=True, 
                  headwidth=6, 
                  headlength=7)
        
    if cmap is not None:
        
        arg_f = np.mod(np.angle(f),2*np.pi)
        
        norm = matplotlib.colors.Normalize(vmin=0,vmax=2*np.pi)
        c_m = cmap # "twilight", "hsv", ...
        s_m = matplotlib.cm.ScalarMappable(cmap=c_m, norm=norm)
        s_m.set_array([])
        
        # Plot the vectors.
        ax.quiver(x, y, np.real(f), np.imag(f), arg_f, 
                  cmap=cmap,
                  pivot="middle",
                  headwidth=6, 
                  headlength=7)

        # Add a colorbar.
        cbar = plt.colorbar(s_m, ticks=[0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], pad=0.1)
        cbar.ax.set_yticklabels(["$0$", "$\\frac{\\pi}{2}$", "$\\pi$", "$\\frac{3\\pi}{2}$", "$2\\pi$"], fontsize=16)
    
    # Axis labels.
    ax.set_xlabel("Re", fontsize=14)
    ax.set_ylabel("Im", fontsize=14)
    
    # Grid and title
    ax.grid(grid)
    
    if title is not None:
        ax.set_title(title, fontsize=18, pad=20, usetex=False)
        
    # Dark background.
    if dark_background == True:
        ax.set_facecolor('black')
        
    plt.tight_layout()
    plt.show()



def complex_streamplot(x, y, f,
                       figsize=(12,8),
                       title=None,
                       grid=False,
                       color="blue",
                       cmap=None,
                       dark_background=False,
                       mod_as_linewidths=False,
                       density=1,
                       scatterpoints=[],
                       pointsize=70,  # The default is 20.
                       pointcolor="black", 
                       pointalpha=1, 
                       pointedgecolors="black", 
                       pointlw=1.5, 
                       pointmarker="o"):
    
    """
    Plot the complex function f as a 2D streamplot.
    It is similar to a vector field plot where the 
    plotted vectors have the form (Real(f), Imag(f)). 
    In polar form, the argument of the function f can
    be represented using a colormap in addition to the
    already visualized orientation of the stream vectors.
    The module of f can be represented as the thickness 
    of the lines of the stream if mod_as_linewidths is 
    set to be True. Scatterpoints can also be added to
    the plot.

    Arguments:

        x, y :: 2D arrays. They represent the 2D plotting 
                space.

        f :: 2D numpy array of complex numbers. Evaluated 
             function to be plotted.

        dark_background :: Boolean. If True, sets the axis 
                           background color to black.

        mod_as_linewidths :: Boolean. If True, the module 
                             of f is represented as the 
                             thickness of the lines of the 
                             stream.

        figsize, title, grid, color, cmap and density are 
        parameters for Matplotlib.

        scatterpoints :: List of complex numbers. The points 
                         contained in this list will be 
                         plotted as scatterpoints.

        pointsize, pointcolor, pointalpha, pointedgecolors, 
        pointlw and pointmarker are parameters defining the 
        properties of the scatter points. These parameters 
        are passed to Matplotlib.
    """

    # Create the figure and the axis.
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.set_aspect("equal")
    
    # Colormap.
    if cmap is None:
        
        if mod_as_linewidths == False:
            
            ax.streamplot(x, y, np.real(f), np.imag(f), color=color, density=density)
            
        if mod_as_linewidths == True:
            
            abs_f = np.log2(np.abs(f)+1)
            abs_f = abs_f/np.max(abs_f)
            ax.streamplot(x, y, np.real(f), np.imag(f), color=color, linewidth=7*abs_f, density=density)
        
    if cmap is not None:
        
        arg_f = np.mod(np.angle(f),2*np.pi)
        
        norm = matplotlib.colors.Normalize(vmin=0,vmax=2*np.pi)
        c_m = cmap #twilight, hsv
        s_m = matplotlib.cm.ScalarMappable(cmap=c_m, norm=norm)
        s_m.set_array([])
        
        if mod_as_linewidths == False:
            
            ax.streamplot(x, y, np.real(f), np.imag(f), color=arg_f, cmap=cmap, density=density)
            
        if mod_as_linewidths == True:
            
            abs_f = np.log2(np.abs(f)+1)
            abs_f = abs_f/np.max(abs_f)
            ax.streamplot(x, y, np.real(f), np.imag(f), color=arg_f, cmap=cmap, linewidth=7*abs_f, density=density)

        cbar = plt.colorbar(s_m, ticks=[0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], pad=0.1)
        cbar.ax.set_yticklabels(["$0$", "$\\frac{\\pi}{2}$", "$\\pi$", "$\\frac{3\\pi}{2}$", "$2\\pi$"], fontsize=16)
        
    # Plot the scatterpoints. 
    if len(scatterpoints) != 0:
        scatterpoints = np.asarray(scatterpoints)
        ax.scatter(scatterpoints.real, scatterpoints.imag, 
                   s=pointsize, 
                   color=pointcolor, 
                   alpha=pointalpha, 
                   edgecolors=pointedgecolors, 
                   linewidths=pointlw, 
                   marker=pointmarker,
                   zorder=100)
    
    # Axis labels.
    ax.set_xlabel("Re", fontsize=14)
    ax.set_ylabel("Im", fontsize=14)
    
    # Grid and title
    ax.grid(grid)
    
    # Title.
    if title is not None:
        ax.set_title(title, fontsize=18, pad=20, usetex=False)
    
    # Set the axis background color to black.
    if dark_background == True:
        ax.set_facecolor('black')
        
    plt.tight_layout()
    plt.show()



def complex_contour(x, y, f, 
                    mode="real",
                    figsize=(8,8),
                    levels=20,
                    xlabel="Re", 
                    ylabel="Im",
                    clabels=True,
                    title=None,
                    usetex=False,
                    grid=False,
                    axis=True,
                    cmap="viridis",
                    ls="solid",
                    lw=1,
                    scatterpoints=[],
                    pointsize=70,  # The default is 20.
                    pointcolor="black", 
                    pointalpha=1, 
                    pointedgecolors="black", 
                    pointlw=1.5, 
                    pointmarker="o",
                    dark_background=False,
                    imshow=False,
                    imcmap="coolwarm"):
        
    """
    Plot either the real or the imaginary part of f (or 
    both) as a contour plot. Scatterpoints can also be 
    added to the plot.

    Arguments:

        x, y :: 2D arrays. They represent the 2D plotting 
                space.

        f :: 2D numpy array of complex numbers. Evaluated 
             function to be plotted.

        mode :: "real", "imag", "modulus" or "both". 
                Choose between plotting the contour of 
                either the real part of f, the imaginary 
                part or both.

        figsize, levels, xlabel, ylabel, clabels, title, 
        usetex, grid, axis, cmap ls and lw are parameters 
        for Matplotlib. levels can be either a list or an 
        integer.

        scatterpoints :: List of complex numbers. The points 
                         contained in this list will be 
                         plotted as scatterpoints.

        pointsize, pointcolor, pointalpha, pointedgecolors, 
        pointlw and pointmarker are parameters defining the 
        properties of the scatter points. These parameters 
        are passed to Matplotlib.

        dark_background :: Boolean. If True, sets the axis 
                           background color to black.

        imshow :: Boolean. If True, shows the module of f as 
                  an imshow plot. Only works if mode!="both".

        imcmap :: String. Colormap for the imshow plot (only 
                  used if imshow=True and mode!="both").
    """
    
    # Get the limit for the plot.
    lim = np.max([x, y])  

    # Create the figure and the axis.
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.set_aspect("equal")
    ax.set_xlabel(xlabel, fontsize=14, usetex=usetex)
    ax.set_ylabel(ylabel, fontsize=14, usetex=usetex)
    
    # Limit corrections
    ax.set_xlim((-lim,lim))
    ax.set_ylim((-lim,lim))
    
    # Decide whether the subplot's axis is shown or not.
    ax.axis(axis)
    # If grid=True sets the grid to be displayed.
    ax.grid(grid)
    
    # Set the default title of the plot.
    if title is None:
        if mode == "real":
            title = "Contour of Re $f(z)$"
        if mode == "imag":
            title = "Contour of Im $f(z)$"
        if mode == "both":
            title = "Contour of Re $f(z)$ & Im $f(z)$"
        if mode == "modulus":
            title = "Contour of $|f(z)|$"
    
    ax.set_title(title, fontsize=18, pad=20, usetex=usetex)
    
    # Mode of the plot.
    if mode != "both":
        
        if mode == "imag":
            f2 = f.imag    
        if mode == "real":
            f2 = f.real
        if mode == "modulus":
            f2 = np.abs(f)
        
        # Plot the contourf.
        cont = ax.contour(x, y, f2, levels=levels, linestyles=ls, linewidths=lw, cmap=cmap)
        # cont = ax.contourf(x, y, np.array(np.abs(f2), dtype=float), levels=levels, cmap=cmap)
        
        # Labels for the contour lines.
        if clabels == True:
            ax.clabel(cont, fontsize=9, inline=1)
            
        if imshow == True:
            ax.imshow(np.array(f2, dtype=float), cmap=imcmap, extent=[-lim,lim,-lim,lim], interpolation="none", origin="lower") #cmap="GnBu" #force float type in f
        
    if mode == "both":
        
        cont_re = ax.contour(x, y, f.real, levels=levels, linestyles=ls, linewidths=lw, colors="C3") #C0 = default red
        if clabels == True:
            ax.clabel(cont_re, fontsize=9, inline=1)
        
        cont_im = ax.contour(x, y, f.imag, levels=levels, linestyles=ls, linewidths=lw, colors="C0") #C3 = default blue
        if clabels == True:
            ax.clabel(cont_im, fontsize=9, inline=1)

        # Add a legend to the contour plot.
        le_re, _ = cont_re.legend_elements()
        le_im, _ = cont_im.legend_elements()
        ax.legend([le_re[0], le_im[0]], ["Re $f(z)$", "Im $f(z)$"], loc="best")
    
    # Plot the scatterpoints.
    if len(scatterpoints) != 0:
        scatterpoints = np.asarray(scatterpoints)
        ax.scatter(scatterpoints.real, scatterpoints.imag, 
                   s=pointsize, 
                   color=pointcolor, 
                   alpha=pointalpha, 
                   edgecolors=pointedgecolors, 
                   linewidths=pointlw, 
                   marker=pointmarker,
                   zorder=100)
    
    # Dark background.
    if dark_background == True:
        ax.set_facecolor('black')
        
    plt.tight_layout()
    plt.show()



"""
File: json_editor.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Manipulate the results.json file
"""
import datetime
import json
from pathlib import Path
import shutil


if __name__ == "__main__":
    # input results.json
    p_in = Path("/src/experiments/results.json")
    # create backup in results_backup folder
    p_backup = Path(
        f"/src/experiments/results_backups/results_backup_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.json")
    p_backup.parent.mkdir(exist_ok=True, parents=True)
    shutil.copy(p_in, p_backup)
    # read original json
    with open(p_in, "r", encoding="utf-8") as f:
        results = json.load(f)

    # modify results
    results_modified = []
    # iterate through all results (all experiments as list entries in results.json file)
    for result in results:
        # filter as desired. Everything not appended to results_modified is deleted
        #if not result["dataset_name"].startswith("ff_"):
        #if not (result["model_name"] == "CVKANWrapper" and result["dataset_name"].startswith("ff_")):
        #if not (result["model_name"] == "CVKANWrapper" and result["dataset_name"] == "knot_complex" and result["use_norm"][0] in ["batchnorm", "batchnormvar"]):
        if not (result["model_name"]=="CVKANWrapper" and result["dataset_name"]=="ph_holo_c_100k"):
            results_modified.append(result)
        else:
            print("deleting", result)
    # overwrite modified results.json
    with open(p_in, "w", encoding="utf-8") as f:
        json.dump(results_modified, f)



"""
File: visualizations.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Create the visualizations and figures used in our paper
"""
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch
import torchmetrics
from pathlib import Path

from src.cvkan.experiments.knot_dataset import load_knot_dataset
from src.cvkan.models.CVKAN import Norms
from src.cvkan.models.wrapper.CVKANWrapper import CVKANWrapper
from src.cvkan.train.train_loop import train_kans
from src.cvkan.utils.dataloading.create_complex_dataset import create_complex_dataset
from src.cvkan.utils.dataloading.csv_dataloader import CSVDataset
from src.cvkan.utils.eval_model import plot_confusion_matrix
from src.cvkan.utils.explain_kan import KANExplainer
from src.cvkan.utils.loss_functions import MSE, MAE
from src.cvkan.utils.plotting.plot_kan import KANPlot
import src.cvkan.utils.plotting.cplotting_tools as cplt
matplotlib.use('TkAgg')
plt.rcParams.update({
    "text.usetex": True,
    "font.family": "Helvetica",
    "text.latex.preamble": r"\usepackage{amsfonts}"
})
vis_root_path = Path(__file__).parent
def rbfs():
    """Create figure of real-valued RBFs forming learnable activation function (images/RBF_weighted_sum.svg)"""
    font = {'family': 'normal',
            'weight': 'bold',
            'size': 22}

    matplotlib.rc('font', **font)
    plt.figure(figsize=(12,10))
    h = 1
    grid_points = np.array([[-2], [0], [2]])
    weights = [[1.3],[-1.2],[0.7]]
    xs = np.arange(-5, 5, 0.01)
    rbfs = np.exp(-(xs - grid_points)**2 / (2*h**2))
    rbfs = weights * rbfs
    plt.plot(xs, rbfs[0,:], color="red", linewidth=0.5, label="RBF at $g_0=-2$ with $w_0=1.3$")
    plt.plot(xs, rbfs[1, :], color="blue", linewidth=0.5, label="RBF at $g_1=0$ with $w_1=-1.2$")
    plt.plot(xs, rbfs[2, :], color="green", linewidth=0.5, label="RBF at $g_2=2$ with $w_2=0.7$")
    plt.plot(xs, np.sum(rbfs, axis=0), color="black", linewidth=3, label="$\Phi(x)$ result of weighted summation")
    plt.plot([-2,-2],[-2.5,1.5],color="black", linewidth=1, linestyle="dashed")
    plt.plot([2, 2], [-2.5, 1.5], color="black", linewidth=1, linestyle="dashed")
    plt.xlim([-5, +5])
    plt.ylim([-2.5,1.5])
    plt.ylabel("y")
    plt.xlabel("x")
    plt.legend()

    #plt.show()
    plt.savefig(vis_root_path / '/RBF_weighted_sum.svg', bbox_inches='tight', transparent=True)
def rbf_hats():
    """Create figure of complex-valued RBFs (images/CRBF_hats.pdf)"""
    font = {'family': 'normal',
            'weight': 'bold',
            'size': 22}

    matplotlib.rc('font', **font)
    plt.figure(figsize=(12,10))
    ax = plt.axes(projection='3d')
    h = 1
    num_grids = 3
    grid_min, grid_max, num_pts = -2, 2, 200

    # create meshgrid for real and imaginary parts of gridpoints
    reals_g, imags_g = np.meshgrid(np.linspace(grid_min, grid_max, num_grids),
                               np.linspace(grid_min, grid_max, num_grids))
    grid = torch.complex(torch.from_numpy(reals_g), torch.from_numpy(imags_g)).to(torch.complex64)

    # create meshgrid for real and imaginary parts of inputs
    reals_i, imags_i = np.meshgrid(np.linspace(-5, 5, num_pts),
                               np.linspace(-5, 5, num_pts))
    inputs = torch.complex(torch.from_numpy(reals_i), torch.from_numpy(imags_i)).to(torch.complex64)
    inputs = inputs.flatten()

    # grid has shape num_grids x num_grids
    # transform x to shape BATCH x Input-Dim x num_grids x num_grids
    x = inputs.unsqueeze(-1).unsqueeze(-1).expand(
        inputs.shape + (num_grids, num_grids))
    # apply RBF
    result = torch.exp(-(torch.abs(x - grid)) ** 2 / (2*h**2))
    result = result.reshape(num_pts, num_pts, num_grids, num_grids)
    # colors for the 9 RBF spikes / hats
    colors = [["tab:blue", "tab:orange", "tab:green"], ["tab:red", "tab:purple", "tab:brown"], ["tab:pink", "tab:olive", "tab:cyan"]]

    # iterate over first grid dimension
    for u in (range(num_grids)):
        # iterate over second grid dimension
        for v in (range(num_grids)):
            # if function value is kess than 0.05, then set it to NaN (limit spikes / hats to certain locations;
            # otherwise the functions would continue forever and all overlap weirdly in the plot)
            for i1 in range(num_pts):
                for i2 in range(num_pts):
                    if result[i1, i2, u, v] <= 0.05:
                        result[i1, i2, u, v] = torch.nan

            # plot spike / hat for gridpoint (u,v)
            ax.plot_surface(reals_i, imags_i, result[:,:,u,v], color=colors[u][v], alpha=0.7)
    # create black plane at zero
    ax.plot_surface(reals_i, imags_i, np.zeros_like(reals_i), color="black", alpha=0.7)
    ax.set_xlabel("$\Re(x)$", labelpad=10)
    ax.set_ylabel("$\Im(x)$", labelpad=10)
    ax.set_zlabel("$RBF_\mathbb{C}(x)$", labelpad=10)
    #plt.show()
    plt.savefig(vis_root_path / 'CRBF_hats.pdf', transparent=True, bbox_inches='tight', pad_inches=0.0)

def plot_kan_square_square():
    """
    Plot model for CVKAN trained on $f(z1, z2) = (z1^2 + z2^2)^2$.
    Produced plot is images/cvkan_sqsq_plot.png
    """
    font = {'family': 'normal',
            'weight': 'bold',
            'size': 22}

    matplotlib.rc('font', **font)
    mse_loss = MSE()
    mae_loss = MAE()
    # define MSE and MAE losses
    loss_fns = dict()
    loss_fns["mse"] = mse_loss
    loss_fns["mae"] = mae_loss

    # lambda for $f(z1, z2) = (z1^2 + z2^2)^2$
    sqsq = lambda x: (x[:, [0]] ** 2 + x[:, [1]] ** 2)**2
    # create dataset
    dataset = create_complex_dataset(sqsq, 2, [-1, 1], train_num=10000, test_num=5000, device="cuda")
    dataset = CSVDataset(dataset, input_vars=["$z_1$", "$z_2$"], output_vars=["$(z_1^2 + z_2^2)^2$"], categorical_vars=[])
    # create model and explainer
    model = CVKANWrapper(layers_hidden=[2,1,1], num_grids=8, rho=1, use_norm=Norms.NoNorm)
    model.to("cuda")
    kan_explainer = KANExplainer(model, samples=dataset.data["test_input"].to("cuda"), method="pykan")
    # train the model to sufficient precision (5000 epochs)
    train_kans(model, dataset=dataset, loss_fn_backprop=mse_loss, loss_fns=loss_fns, device="cuda", batch_size=2000,logging_interval=100,add_softmax_lastlayer=False,epochs=5000,last_layer_output_real=False, sparsify=True, kan_explainer=kan_explainer)

    # plot the model
    kan_plotter = KANPlot(model,kan_explainer=kan_explainer, input_featurenames=dataset.input_varnames, output_names=dataset.output_varnames, complex_valued=True, plot_options=None)
    kan_plotter.plot_all()
def plot_kan_knot():
    """
    Plot model for CVKAN trained on knot dataset.
    Produced plot is images/cvkan_knot_relevances.png
    as well as images/cvkan_knot_confusionmatrix.svg
    """
    font = {'family': 'normal',
            'weight': 'bold',
            'size': 12}
    matplotlib.rc('font', **font)
    # define CE-Loss and accuracy
    crossentropy_loss = torch.nn.CrossEntropyLoss()
    loss_fns = dict()
    loss_fns["cross_entropy"] = crossentropy_loss
    loss_fns["accuracy"] = torchmetrics.Accuracy(task="multiclass", num_classes=14).to("cuda")
    # load knot dataset
    dataset = load_knot_dataset(complex_dataset=True)
    # create model and explainer
    model = CVKANWrapper(layers_hidden=[15,1,14], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
    model.to("cuda")
    kan_explainer = KANExplainer(model, samples=dataset.data["test_input"], method="pykan")
    # train CVKAN model
    train_kans(model, dataset=dataset, loss_fn_backprop=crossentropy_loss, loss_fns=loss_fns, device="cuda", batch_size=10000,
               logging_interval=10, add_softmax_lastlayer=True, epochs=200, last_layer_output_real=True, sparsify=True, kan_explainer=kan_explainer)
    # plot the model
    kan_plotter = KANPlot(model, kan_explainer=kan_explainer, input_featurenames=dataset.input_varnames,
                          output_names=dataset.output_varnames, complex_valued=True, plot_options=None)
    kan_plotter.plot_all()
    model.to("cuda")
    # plot confusionmatrix. Labelmapping converts class indices to class names. Highly underrepresented labels
    # have been removed as they are very unlikely (but not impossible) to show up in test split
    labelmapping = [-12, -10, -8, -6,-4,-2,0,2,4,6,8,10,12]  # removed 14
    plot_confusion_matrix(pred=model(dataset.data["test_input"].to("cuda")), gt=dataset.data["test_label"].to("cuda"), labelmapping=labelmapping)
def plot_ideal_zsquared():
    """
    Plot an ideal $f(z)=z^2$ function
    """
    font = {'family': 'normal',
            'weight': 'bold',
            'size': 24}
    matplotlib.rc('font', **font)
    # meshgrid for real and imaginary parts
    xs, ys = np.meshgrid(np.linspace(-2, 2, 100), np.linspace(-2, 2, 100))
    # convert to complex numbers
    zs = xs + 1j * ys
    # square the complex-valued numbers
    f = zs ** 2
    # plot it
    cplt.complex_plot3D(xs, ys, f, fontsize=24)
if __name__ == '__main__':
    #rbfs()
    rbf_hats()
    #plot_kan_square_square()
    #plot_ideal_zsquared()
    #plot_kan_knot()



"""
File: knot_dataset.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Experiments on the Knot Dataset (download dataset.csv from google storage
             gs://maths_conjectures/knot_theory/knot_theory_invariants.csv
             and also see their Jupyter Notebook:
             https://github.com/google-deepmind/mathematics_conjectures/blob/main/knot_theory.ipynb
             )
"""

from pathlib import Path
import pandas as pd
import torch
import torchmetrics

from ..experiments.run_crossval import run_crossval
from ..models.CVKAN import Norms
from ..models.FastKAN import FastKAN
from ..models.wrapper.CVKANWrapper import CVKANWrapper
from ..models.wrapper.PyKANWrapper import PyKANWrapper
from ..utils.dataloading.csv_dataloader import CSVDataset

def load_knot_dataset(input_filename=Path("/home/m_wolf37/Datasets/knot_theory_invariants.csv"), train_test_split="70:30", complex_dataset=False):
    """
    Load Knot Dataset from .csv file
    :param input_filename: Path to dataset .csv file
    :param train_test_split: String specifying the share of train and test split as 'trainpercent:testpercent'
    :param complex_dataset: Whether return should be a complex-valued dataset (True) or real-valued (False)
    :return: Complex-valued or real-valued dataset
    """
    full_df = pd.read_csv(input_filename)
    full_df.drop(full_df.columns[0], axis=1, inplace=True)  # drop first column (id)
    # extract column names for input and output
    input_vars = list(full_df.columns[0:len(full_df.columns) - 1])
    # varnames for complex-valued dataset (abbreviated). Note that this list contains imag and real parts as
    # one single complex-valued variable
    input_vars_short_complex = ["adjoint_torsion_deg", "torsion_deg", "short_geodesic", "inject_rad", "chern_simons", "cusp_vol", "longit_translat", "merid_translat_c", "volume", "sym_0", "sym_d3", "sym_d4", "sym_d6", "sym_d8", "sym_Z/2+Z/2"]
    output_var = [full_df.columns[len(full_df.columns) - 1]]
    # load dataset (real valued)
    dataset = CSVDataset(full_df, input_vars=input_vars, output_vars=output_var, categorical_vars=output_var, train_test=train_test_split)
    # normalize dataset (to our fixed grid range of [-2, 2])
    dataset.normalize()

    if not complex_dataset:  # if interested in complex-valued knot dataset
        return dataset
    # otherwise: use this dataset to construct a complex-valued dataset
    num_train, num_test = dataset.data["train_input"].shape[0], dataset.data["test_input"].shape[0]
    num_complex_input_vars = len(dataset.input_varnames) - 2  # -2 because there are already 2 complex numbers in the dataset (split in Re and Im)
    dataset_complex = dict()
    dataset_complex["train_input"] = torch.zeros((num_train, num_complex_input_vars), dtype=torch.complex64)
    dataset_complex["test_input"] = torch.zeros((num_test, num_complex_input_vars), dtype=torch.complex64)

    # copy input features into complex dataset setting imaginary part to zero (if it doesnt exist)
    for idx_orig, idx_complex in [(0,0), (1,1), (4,3), (5,4), (6,5), (7,6), (10,8), (11,9), (12,10), (13,11), (14,12), (15,13), (16,14)]:
        print(idx_orig, idx_complex, dataset.data["train_input"].shape)
        dataset_complex["train_input"][:,idx_complex] = torch.complex(dataset.data["train_input"][:,idx_orig], torch.zeros_like(dataset.data["train_input"][:,idx_orig]))
        dataset_complex["test_input"][:,idx_complex] = torch.complex(dataset.data["test_input"][:,idx_orig], torch.zeros_like(dataset.data["test_input"][:,idx_orig]))

    # make complex number for "short geodesic"
    dataset_complex["train_input"][:, 2] = torch.complex(dataset.data["train_input"][:, 2], dataset.data["train_input"][:, 3])
    dataset_complex["test_input"][:, 2] = torch.complex(dataset.data["test_input"][:, 2], dataset.data["test_input"][:, 3])
    # make complex number for "meridinal translation" (in CSV imag and real are swapped...)
    dataset_complex["train_input"][:, 7] = torch.complex(dataset.data["train_input"][:, 9], dataset.data["train_input"][:, 8])
    dataset_complex["test_input"][:, 7] = torch.complex(dataset.data["test_input"][:, 9], dataset.data["test_input"][:, 8])

    # copy labels
    dataset_complex["train_label"] = dataset.data["train_label"]
    dataset_complex["test_label"] = dataset.data["test_label"]

    # change input vars
    input_vars[2] = "short_geodesic_complex"
    input_vars[8] = "meridinal_translation_complex"
    del input_vars[9]
    del input_vars[3]
    dataset_complex = CSVDataset(dataset_complex, input_vars=input_vars, output_vars=["c" + str(i) for i in range(14)], categorical_vars=[])

    dataset_complex.num_classes = dataset.num_classes
    dataset_complex.input_varnames = input_vars_short_complex
    return dataset_complex
def run_experiments():
    """Run the experiments on the Knot Dataset"""
    _DEVICE = torch.device("cuda")
    # load knot complex and real-valued with 100% train split
    # use 100% train split because run_crossval expects it this way and splits it into 5 non-overlapping folds
    knot_dataset_complex = load_knot_dataset(train_test_split="100:0", complex_dataset=True)
    knot_dataset_real = load_knot_dataset(train_test_split="100:0", complex_dataset=False)
    in_features_real = len(knot_dataset_real.input_varnames)
    in_features_complex = len(knot_dataset_complex.input_varnames)
    num_classes = len(knot_dataset_complex.output_varnames)

    crossentropy_loss = torch.nn.CrossEntropyLoss()

    loss_fns = dict()
    loss_fns["cross_entropy"] = crossentropy_loss
    loss_fns["accuracy"] = torchmetrics.Accuracy(task="multiclass", num_classes=num_classes).to(_DEVICE)

    # Experiments for FastKAN and pyKAN
    for arch, use_batchnorm, num_grids in [([in_features_real, 1, num_classes], True, 8), ([in_features_real, 1, num_classes], True, 64), ([in_features_real, 2, num_classes], True, 8), ([in_features_real, 2, num_classes], True, 64)]:
        fastkan = FastKAN(layers_hidden=arch,
                          num_grids=num_grids, use_batchnorm=use_batchnorm, grid_mins=-2, grid_maxs=2)
        run_crossval(fastkan, knot_dataset_real, dataset_name="knot_r", loss_fn_backprop=crossentropy_loss, loss_fns=loss_fns, device=_DEVICE,
                     batch_size=10000, logging_interval=50, add_softmax_lastlayer=True, epochs=200, convert_model_output_to_real=False)
        pykan = PyKANWrapper(layers_hidden=arch, num_grids=num_grids, update_grid=use_batchnorm, device=_DEVICE, grid_range=[-2,2])
        run_crossval(pykan, knot_dataset_real, dataset_name="knot_r", loss_fn_backprop=loss_fns["cross_entropy"],
                     loss_fns=loss_fns,
                     batch_size=-1, add_softmax_lastlayer=True, epochs=200, convert_model_output_to_real=False, device=torch.device("cuda"))

    # Experiments for our CVKAN
    arch = [in_features_complex, 2, num_classes]
    num_grids = 8
    for batchnorm in [Norms.BatchNorm, Norms.BatchNormVar, Norms.BatchNormNaiv, Norms.NoNorm]:
        for zsilu_type in ["complex_weight", "real_weights"]:
            cvkan = CVKANWrapper(layers_hidden=arch, num_grids=num_grids, rho=1, use_norm=batchnorm, grid_mins=-2, grid_maxs=2, zsilu_type=zsilu_type)
            run_crossval(cvkan, knot_dataset_complex, dataset_name="knot_c", loss_fn_backprop=crossentropy_loss,
                         loss_fns=loss_fns, device=_DEVICE,
                         batch_size=10000, logging_interval=50, add_softmax_lastlayer=True, epochs=200, convert_model_output_to_real=True)


def train_knot_feature_subset():
    """Training on the Knot Dataset on the most important 3 or 7 features only
    as well as on the inverse set of every feature except the most important 3 or 7"""
    # use 100% train split because run_crossval expects it this way and splits it into 5 non-overlapping folds
    knot_dataset = load_knot_dataset(train_test_split="100:0", complex_dataset=True)
    # indices for features to train on
    indices_big = [1,2,3,5,6,7,8]
    indices_small = [2,6,7]
    indices_big_inverse = [0,4,9,10,11,12,13,14]
    indices_small_inverse = [0,1,3,4,5,8,9,10,11,12,13,14]

    num_classes = 14
    _DEVICE = torch.device("cuda")
    crossentropy_loss = torch.nn.CrossEntropyLoss()

    loss_fns = dict()
    loss_fns["cross_entropy"] = crossentropy_loss
    loss_fns["accuracy"] = torchmetrics.Accuracy(task="multiclass", num_classes=num_classes).to(_DEVICE)
    # train only on a specific subset of input-features
    for indices in [indices_small, indices_big, indices_small_inverse, indices_big_inverse]:
        cvkan = CVKANWrapper(layers_hidden=[len(indices), 1, 14], num_grids=8, rho=1, use_norm=Norms.BatchNorm, grid_mins=-2, grid_maxs=2, zsilu_type="complex_weight")
        train_input = knot_dataset.data["train_input"][:,indices]
        # reduce dataset to the specified subset of features
        knot_dataset_complex_reduced = dict()
        knot_dataset_complex_reduced["train_input"] = train_input
        knot_dataset_complex_reduced["train_label"] = knot_dataset.data["train_label"]
        knot_dataset_complex_reduced["test_label"] = knot_dataset.data["test_label"]
        knot_dataset_complex_reduced["test_input"] = knot_dataset.data["test_label"]
        knot_dataset_complex_reduced = CSVDataset(knot_dataset_complex_reduced, input_vars=[knot_dataset.input_varnames[i] for i in indices], output_vars=knot_dataset.output_varnames, categorical_vars=[])

        run_crossval(cvkan, knot_dataset_complex_reduced, dataset_name="knot_c_"+str(indices), loss_fn_backprop=crossentropy_loss,
                     loss_fns=loss_fns, device=_DEVICE,
                     batch_size=10000, logging_interval=50, add_softmax_lastlayer=True, epochs=200,
                     convert_model_output_to_real=True)

if __name__ == "__main__":
    #run_experiments()
    train_knot_feature_subset()



"""
File: WrapperTemplate.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Interface for KANs. To make them work with our KanPlotter and KanExplainer all the methods need
             to be supported by the specific KAN variant/wrapper.
"""

from abc import abstractmethod

import numpy as np
import torch


class WrapperTemplate:
    def __init__(self):
        pass
    @abstractmethod
    def plot_curve(self, l_i_j_index: tuple, num_pts = 1000) -> (np.ndarray, tuple[np.ndarray]):
        """
        Returns the learned Activation Function with index (l,i,j) connecting Node i in Layer l to Node j in Layer l+1
        :param l_i_j_index: Tuple for Index (l,i,j)
        :param num_pts: Number of points for sampling
        :return: Tuple (xs, (ys_base, ys_splines, ys_sum_total)) with input Values and output Values per component
        xs are input values, ys_base the base activation (i.e. SiLU), ys_splines the individual Splines/RBFs/... and
        ys_sum_total is the sum of everything (so the final learned function)
        """
        raise NotImplementedError
    @abstractmethod
    def get_edge_activations(self, l_i_j_index: tuple):
        """
        Returns Activations of the Edge with Index (l,i,j)
        :param l_i_j_index: Tuple (l,i,j)
        :return: Activations across all Samples for given Edge provided in 'gather_activations(...)' with Dimension [N]
        """
        raise NotImplementedError
    @abstractmethod
    def get_node_activations(self, l_i_index: tuple):
        """
        Returns Activations of the Node with Index (l,i)
        :param l_i_index: Tuple (l,i)
        :return: Activations across all Samples for given Node provided in 'gather_activations(...)' with Dimension [N]
        """
        raise NotImplementedError
    @abstractmethod
    def get_layersizes(self) -> list:
        """
        Returns the widths of all layers
        :return: List with one value for each layer starting at the bottom (input side) of the Network representing
        layer's width
        """
        raise NotImplementedError
    @abstractmethod
    def get_gridrange(self) -> list:
        """
        Returns the grid-range as a List
        :return: Grid Range
        """
        # ToDo: maybe change this for only one single Activation Function, not for the whole Network/Layer
        raise NotImplementedError
    @abstractmethod
    def get_kan_layers(self) -> torch.nn.ModuleList:
        """
        Returns all Layers as a ModuleList
        :return: ModuleList of all Layers
        """
        raise NotImplementedError
    @abstractmethod
    def gather_activations(self, x):
        """
        Collects the Activations on every Edge and Node for later Usage
        :param x: Input Samples
        """
        raise NotImplementedError
    def get_all_weights(self):
        """
        Returns all weights as a concatenated, flattened Tensor. To be used for Regularization.
        :return: Concatenated, flattened Tensor with all trainable Weights
        """
        raise NotImplementedError


"""
File: ComplexSilu.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Two different variants of complex SiLU
"""
import torch


def complex_silu_complexweight(complex_input):
    """
    Applies SiLU on real and imaginary parts seperately, then combines both into a complex number and returns it
    :param complex_input: Input
    :return: SiLu(Input.real) + i*SiLU(Input.imag)
    """
    return torch.complex(torch.nn.functional.silu(complex_input.real, inplace=False), torch.nn.functional.silu(complex_input.imag, inplace=False))
def complex_silu_realweights(complex_input):
    """
    Calculates SiLU on real and imaginary parts seperately, returns result as a 2-tuple (so we can weight each element
    differently later)
    :param complex_input: Input
    :return: Tuple (SiLU(Input.real), SiLU(Input.imag))
    """
    return (torch.nn.functional.silu(complex_input.real, inplace=False), torch.nn.functional.silu(complex_input.imag, inplace=False))


"""
File: fit_formulas.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Experiments for Function Fitting (physically meaningful circuit & holography
             as well as arbitrary simple formulae)
"""

import argparse
import torch
from kan import create_dataset

from ..experiments.run_crossval import run_crossval
from ..models.CVKAN import Norms
from ..models.FastKAN import FastKAN
from ..models.wrapper.CVKANWrapper import CVKANWrapper
from ..models.wrapper.PyKANWrapper import PyKANWrapper
from ..utils.dataloading.create_complex_dataset import create_complex_dataset
from ..utils.dataloading.csv_dataloader import CSVDataset
from ..utils.loss_functions import MSE, MAE

mse_loss = MSE()
mae_loss = MAE()

loss_fns = dict()
loss_fns["mse"] = mse_loss
loss_fns["mae"] = mae_loss

def convert_complex_dataset_to_real(dataset: CSVDataset):
    """Converts a complex dataset to a real-valued dataset by doubling input and output dimension (one real number for
    real and imaginary part each)"""
    if dataset.categorical_vars:
        raise NotImplementedError("complex dataset can't contain categorical vars!")
    num_vars_complex = len(dataset.input_varnames)
    num_vars_real = 2 * num_vars_complex
    num_outputs_complex = len(dataset.output_varnames)
    num_outputs_real = 2 * num_outputs_complex
    num_samples_train, num_samples_test = dataset.get_train_test_size()

    train_input = torch.zeros((num_samples_train, num_vars_real), dtype=torch.float32)
    test_input = torch.zeros((num_samples_test, num_vars_real), dtype=torch.float32)
    train_label = torch.zeros((num_samples_train, num_outputs_real), dtype=torch.float32)
    test_label = torch.zeros((num_samples_test, num_outputs_real), dtype=torch.float32)

    # make sure varnames are also doubled for later plotting
    input_varnames_real = []
    output_varnames_real = []

    for in_feature in range(num_vars_complex):
        train_input[:, 2*in_feature] = dataset.data["train_input"][:, in_feature].real
        train_input[:, 2*in_feature+1] = dataset.data["train_input"][:, in_feature].imag

        if num_samples_test > 0:
            test_input[:, 2*in_feature] = dataset.data["test_input"][:, in_feature].real
            test_input[:, 2*in_feature+1] = dataset.data["test_input"][:, in_feature].imag

        input_varnames_real.append(dataset.input_varnames[in_feature] + ".real")
        input_varnames_real.append(dataset.input_varnames[in_feature] + ".imag")
    for out_feature in range(num_outputs_complex):
        train_label[:, 2*out_feature] = dataset.data["train_label"][:, out_feature].real
        train_label[:, 2*out_feature+1] = dataset.data["train_label"][:, out_feature].imag

        if num_samples_test > 0:
            test_label[:, 2*out_feature] = dataset.data["test_label"][:, out_feature].real
            test_label[:, 2*out_feature+1] = dataset.data["test_label"][:, out_feature].imag

        output_varnames_real.append(dataset.output_varnames[out_feature] + ".real")
        output_varnames_real.append(dataset.output_varnames[out_feature] + ".imag")
    # build a dictionary out of the now real-valued datapoints
    realdata_dict = dict()
    realdata_dict['train_input'] = train_input
    realdata_dict['train_label'] = train_label
    realdata_dict['test_input'] = test_input
    realdata_dict['test_label'] = test_label

    # create a CVDataset object from this dict
    dataset_real = CSVDataset(realdata_dict, input_vars=input_varnames_real, output_vars=output_varnames_real, categorical_vars=[])
    return dataset_real

def run_experiments_physics(run_dataset, run_model):
    """Main method to run experiments on physically meaningul formula fitting (circuit & holography)"""
    _num_samples = 100000
    _dataset_name_suffix = "_100k"  # differentiate runs on 100k and 5k samples
    loss_fn_backprop = loss_fns["mse"]

    # holography formula
    holography = lambda x: torch.abs(x[:, [0]] + x[:, [1]]) ** 2 * x[:, [2]]
    # only generate train samples and no test samples because run_crossval later splits them into k folds
    dataset_holography_c = create_complex_dataset(holography, ranges=[-2,2], n_var=3, train_num=_num_samples, test_num=0)
    dataset_holography_c = CSVDataset(dataset_holography_c, input_vars=["Er1", "E0", "Er2"], output_vars=["holography"],
                                      categorical_vars=[])
    # create real-valued holography dataset from complex one
    dataset_holography_r = convert_complex_dataset_to_real(dataset_holography_c)

    # circuit dataset formula (complex and real separately, as most variables are not complex to begin with so we can't
    # easily convert complex dataset to real one)
    # variable indices: 0: U_g, 1: R_g, 2: R_l, 3: w, 4: L, 5: C
    circuit_complex_inputs = lambda x: x[:, [0]] / (1 + x[:, [1]].real/x[:, [2]].real - x[:, [3]].real**2 * x[:, [4]].real * x[:, [5]].real + 1j * x[:, [3]].real * (x[:, [4]].real/ x[:, [2]].real + x[:, [1]].real * x[:, [5]].real))
    # variable indices: 0: U_g.real, 1: U_g.imag, 2: R_g, 3: R_l, 4: w, 5: L, 6: C
    circuit_real_inputs_helper = lambda x: (torch.complex(x[:, [0]], x[:, [1]]) /
                                     torch.complex(1+x[:, [2]]/x[:, [3]] - x[:, [4]]**2 * x[:, [5]] * x[:, [6]], x[:, [4]] * (x[:, [5]] / x[:, [3]] + x[:, [2]] * x[:, [6]])))
    circuit_real_inputs = lambda x: torch.stack((circuit_real_inputs_helper(x).real, circuit_real_inputs_helper(x).imag), dim=1).squeeze()
    # only generate train samples and no test samples because run_crossval later splits them into k folds
    dataset_circuit_c = create_complex_dataset(circuit_complex_inputs, ranges=[-2, 2], n_var=6, train_num=_num_samples, test_num=0)
    # zero out imaginary parts of real-input-variables in complex dataset
    for var_idx in [1,2,3,4,5]:
        dataset_circuit_c["train_input"][:, var_idx] = torch.complex(dataset_circuit_c["train_input"][:, var_idx].real, 0*dataset_circuit_c["train_input"][:, var_idx].imag)
    dataset_circuit_c = CSVDataset(dataset_circuit_c, input_vars=["U_g", "R_g", "R_l", "w", "L", "C"], output_vars=["U_{rl}"], categorical_vars=[])
    # only generate train samples and no test samples because run_crossval later splits them into k folds
    dataset_circuit_r = create_dataset(circuit_real_inputs, ranges=[-2, 2], n_var=7, train_num=_num_samples, test_num=0)
    dataset_circuit_r = CSVDataset(dataset_circuit_r, input_vars=["U_g.real","U_g.imag", "R_g", "R_l", "w", "L", "C"],
                                   output_vars=["U_{rl}.real", "U_{rl}.imag"], categorical_vars=[])

    # check which model and dataset to run
    run_models = [False] * 3
    if run_model == "all":
        run_models = [True] * 3
    elif run_model == "pykan":
        run_models[0] = True
    elif run_model == "fastkan":
        run_models[1] = True
    elif run_model == "cvkan":
        run_models[2] = True

    run_datasets = [False] * 2
    if run_dataset == "all":
        run_datasets = [True] * 2
    elif run_dataset == "holography":
        run_datasets[0] = True
    elif run_dataset == "circuit":
        run_datasets[1] = True

    if run_datasets[0]:  # holography
        for arch in [[6,10,5,3,2], [6,1,2], [6,5,2], [6,10,2]]:
            if run_models[0]:
                # important to put batch_size = -1 for PyKAN as they do batching differently (randomly sampling
                # batch_size samples per step)
                pykan = PyKANWrapper(layers_hidden=arch, num_grids=64, update_grid=True, device="cuda")
                run_crossval(pykan, dataset_holography_r, dataset_name="ph_holo_r"+_dataset_name_suffix, loss_fn_backprop=loss_fn_backprop,
                             loss_fns=loss_fns,
                             batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
            if run_models[1]:
                fastkan = FastKAN(layers_hidden=arch, num_grids=64, use_batchnorm=True)
                run_crossval(fastkan, dataset_holography_r, dataset_name="ph_holo_r"+_dataset_name_suffix, loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                             batch_size=10000, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        for arch in [[3,1], [3,1,1], [3,3,1], [3,10,1], [3,10,3,1], [3,10,5,3,1]]:
            if run_models[2]:
                cvkan = CVKANWrapper(layers_hidden=arch, num_grids=8, rho=1, use_norm=Norms.BatchNorm)
                run_crossval(cvkan, dataset_holography_c, dataset_name="ph_holo_c"+_dataset_name_suffix, loss_fn_backprop=loss_fn_backprop,
                             loss_fns=loss_fns,
                             batch_size=10000, add_softmax_lastlayer=False, epochs=1000,
                             convert_model_output_to_real=False)
    if run_datasets[1]:  # circuit
        for arch in [[7,1,2], [7,5,2], [7,10,2], [7,10,5,3,2]]:
            if run_models[0]:
                # important to put batch_size = -1 for PyKAN as they do batching differently (randomly sampling
                # batch_size samples per step)
                pykan = PyKANWrapper(layers_hidden=arch, num_grids=64, update_grid=True, device="cuda")
                run_crossval(pykan, dataset_circuit_r, dataset_name="ph_circuit_r"+_dataset_name_suffix, loss_fn_backprop=loss_fn_backprop,
                             loss_fns=loss_fns,
                             batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
            if run_models[1]:
                fastkan = FastKAN(layers_hidden=arch, num_grids=64, use_batchnorm=True)
                run_crossval(fastkan, dataset_circuit_r, dataset_name="ph_circuit_r"+_dataset_name_suffix, loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                             batch_size=10000, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        for arch in [[6,1], [6,1,1], [6,3,1], [6,10,1], [6,10,3,1], [6,10,5,3,1]]:
            if run_models[2]:
                cvkan = CVKANWrapper(layers_hidden=arch, num_grids=8, rho=1, use_norm=Norms.BatchNorm)
                run_crossval(cvkan, dataset_circuit_c, dataset_name="ph_circuit_c"+_dataset_name_suffix, loss_fn_backprop=loss_fn_backprop,
                             loss_fns=loss_fns,
                             batch_size=10000, add_softmax_lastlayer=False, epochs=1000,
                             convert_model_output_to_real=False)




def run_experiments_funcfitting(run_dataset = "all", run_model="all"):
    """Main method to run experiments on arbitrary simple formula fitting (z^2, sin(z), z_1*z_2, (z_1^2 + z_2^2)^2)"""
    loss_fn_backprop = loss_fns["mse"]
    sq = lambda x: x[:, [0]]**2
    sqsq = lambda x: ((x[:, [0]]) ** 2 + x[:, [1]] ** 2) ** 2
    mult = lambda x: (x[:, [0]]) * x[:, [1]]
    sinus = lambda x: torch.sin(x[:,[0]])
    run_datasets = [False] * 4
    if run_dataset == "all":
        run_datasets = [True] * 4
    elif run_dataset == "square":
        run_datasets[0] = True
    elif run_dataset == "squaresquare":
        run_datasets[1] = True
    elif run_dataset == "mult":
        run_datasets[2] = True
    elif run_dataset == "sinus":
        run_datasets[3] = True

    run_models = [False] * 3
    if run_model == "all":
        run_models = [True] * 3
    elif run_model == "pykan":
        run_models[0] = True
    elif run_model == "fastkan":
        run_models[1] = True
    elif run_model == "cvkan":
        run_models[2] = True

    # only generate train samples and no test samples because run_crossval later splits them into k folds

    dataset_sq_c = create_complex_dataset(sq, ranges=[-2,2], n_var=1, train_num=5000, test_num=0)
    dataset_sq_c = CSVDataset(dataset_sq_c, input_vars=["z"], output_vars=["z^2"], categorical_vars=[])
    dataset_sq_r = convert_complex_dataset_to_real(dataset_sq_c)

    dataset_sqsq_c = create_complex_dataset(sqsq, ranges=[-2,2], n_var=2, train_num=5000, test_num=0)
    dataset_sqsq_c = CSVDataset(dataset_sqsq_c, input_vars=["z_1", "z_2"], output_vars=["(z_1^2 + z_2^2)^2"], categorical_vars=[])
    dataset_sqsq_r = convert_complex_dataset_to_real(dataset_sqsq_c)

    dataset_mult_c = create_complex_dataset(mult, ranges=[-2, 2], n_var=2, train_num=5000, test_num=0)
    dataset_mult_c = CSVDataset(dataset_mult_c, input_vars=["z_1", "z_2"], output_vars=["z_1 * z_2"],
                            categorical_vars=[])
    dataset_mult_r = convert_complex_dataset_to_real(dataset_mult_c)

    dataset_sin_c = create_complex_dataset(sinus, ranges=[-2,2], n_var=1, train_num=5000, test_num=0)
    dataset_sin_c = CSVDataset(dataset_sin_c, input_vars=["z"], output_vars=["sin(z)"], categorical_vars=[])
    dataset_sin_r = convert_complex_dataset_to_real(dataset_sin_c)

    # Square Dataset = z**2
    if run_datasets[0]:
        if run_models[0]:
            pykan = PyKANWrapper(layers_hidden=[2, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_sq_r, dataset_name="ff_square", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            pykan = PyKANWrapper(layers_hidden=[2, 3, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_sq_r, dataset_name="ff_square", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[1]:
            fastkan = FastKAN(layers_hidden=[2, 2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_sq_r, dataset_name="ff_square", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            fastkan = FastKAN(layers_hidden=[2, 3, 2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_sq_r, dataset_name="ff_square", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[2]:
            cvkan = CVKANWrapper(layers_hidden=[1, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_sq_c, dataset_name="ff_square", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            cvkan = CVKANWrapper(layers_hidden=[1,2, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_sq_c, dataset_name="ff_square", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

    # Square Square Dataset = (z_1**2 + z_2**2)**2
    if run_datasets[1]:
        if run_models[0]:
            pykan = PyKANWrapper(layers_hidden=[4, 2, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_sqsq_r, dataset_name="ff_squaresquare", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            pykan = PyKANWrapper(layers_hidden=[4, 6, 2, 3, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_sqsq_r, dataset_name="ff_squaresquare", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[1]:
            fastkan = FastKAN(layers_hidden=[4, 2, 2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_sqsq_r, dataset_name="ff_squaresquare", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            fastkan = FastKAN(layers_hidden=[4,6,2,3,2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_sqsq_r, dataset_name="ff_squaresquare", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[2]:
            cvkan = CVKANWrapper(layers_hidden=[2,1, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_sqsq_c, dataset_name="ff_squaresquare", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            cvkan = CVKANWrapper(layers_hidden=[2, 4, 2, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_sqsq_c, dataset_name="ff_squaresquare", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

    # Mult Dataset = z_1 * z_2
    if run_datasets[2]:
        if run_models[0]:
            pykan = PyKANWrapper(layers_hidden=[4, 4, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_mult_r, dataset_name="ff_mult", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            pykan = PyKANWrapper(layers_hidden=[4, 8, 4, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_mult_r, dataset_name="ff_mult", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[1]:
            fastkan = FastKAN(layers_hidden=[4, 4, 2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_mult_r, dataset_name="ff_mult", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            # should be optimal fastkan able to compute z1 * z2
            fastkan = FastKAN(layers_hidden=[4,8,4,2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_mult_r, dataset_name="ff_mult", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[2]:
            cvkan = CVKANWrapper(layers_hidden=[2,2, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_mult_c, dataset_name="ff_mult", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            cvkan = CVKANWrapper(layers_hidden=[2, 4, 2, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_mult_c, dataset_name="ff_mult", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

    # sin Dataset = sin(z)
    if run_datasets[3]:
        if run_models[0]:
            pykan = PyKANWrapper(layers_hidden=[2, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_sin_r, dataset_name="ff_sin", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            pykan = PyKANWrapper(layers_hidden=[2, 4, 4, 2], num_grids=64, update_grid=True, device="cuda")
            run_crossval(pykan, dataset_sin_r, dataset_name="ff_sin", loss_fn_backprop=loss_fn_backprop,
                         loss_fns=loss_fns,
                         batch_size=-1, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[1]:
            fastkan = FastKAN(layers_hidden=[2, 2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_sin_r, dataset_name="ff_sin", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            # should be correct network size (sin of complex number = cosh ...)
            fastkan = FastKAN(layers_hidden=[2, 4,4, 2], num_grids=64, use_batchnorm=True)
            run_crossval(fastkan, dataset_sin_r, dataset_name="ff_sin", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)
        if run_models[2]:
            cvkan = CVKANWrapper(layers_hidden=[1, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_sin_c, dataset_name="ff_sin", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)

            cvkan = CVKANWrapper(layers_hidden=[1,2, 1], num_grids=8, rho=1, use_norm=Norms.BatchNorm)
            run_crossval(cvkan, dataset_sin_c, dataset_name="ff_sin", loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                         batch_size=500, add_softmax_lastlayer=False, epochs=1000, convert_model_output_to_real=False)



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', nargs='?', default="all", type=str)
    parser.add_argument('--model', nargs='?', default="all", type=str)
    parser.add_argument("--task", type=str, required=True)

    args = parser.parse_args()
    print("Running Function Fitting Eperiments for Dataset ", args.dataset, " and Model ", args.model, " and Task ", args.task)
    if args.task == "funcfit":
        run_experiments_funcfitting(run_dataset=args.dataset, run_model=args.model)
    if args.task == "physics":
        run_experiments_physics(run_dataset=args.dataset, run_model=args.model)



from .dataloading.create_complex_dataset import create_complex_dataset
from .dataloading.csv_dataloader import CSVDataset
from .plotting.plot_kan import KANPlot
from .explain_kan import KANExplainer

__all__ = [create_complex_dataset, CSVDataset, KANPlot, KANExplainer]


"""
File: train_loop.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Main loop for training all kinds of KANs on any dataset with arbitrary loss functions
"""
import torch
from torch.utils.data import DataLoader

from ..models.wrapper import PyKANWrapper
from ..utils.dataloading.csv_dataloader import CSVDataset
from ..utils.eval_model import eval_model
from ..utils.misc import get_num_parameters


def train_kans(model, dataset: CSVDataset, loss_fn_backprop, loss_fns, device=torch.device("cuda"), epochs=5000,
               batch_size=1000, kan_explainer=None, logging_interval=50, add_softmax_lastlayer=False, last_layer_output_real=True, sparsify=False):

    """
    Train KANs for different datasets. Suitable for complex and real-valued KANs
    :param model: model to train
    :param dataset: dataset to train on (has to be instance of class CSVDataset)
    :param loss_fn_backprop: loss function to be used for backpropagation
    :param loss_fns: dictionary of loss functions to evaluate the model on (key of dict entry should be loss function name)
    :param device: device to train on (default Cuda)
    :param epochs: number of epochs to train for
    :param batch_size: batch size
    :param kan_explainer: kan_explainer to use for regularization of edge's importance scores. Only used if 'sparsify=True'
    :param logging_interval: interval in which results should be logged on console
    :param add_softmax_lastlayer: flag whether softmax should be applied before evaluation the entries of 'loss_fns' dict
    Keep in mind that loss function named 'cross_entropy' will not have SoftMax applied beforehand in any case.
    :param last_layer_output_real: useful for classification tasks where the last layer needs to be real. Simply takes
    the real part of the model's output
    :return: resulting train and test losses at the very end of training. Both of them are a dict with keys representing the
    name of the loss function (same as keys in parameter 'loss_fns')
    """
    # __getitem__ of dataset should give elements from train split
    dataset.set_returnsplit("train")
    # move model to correct device
    model.to(device)
    if type(model) == PyKANWrapper:  # pyKAN (needs to be wrapper, because pykan does not store certain attributes by itself...)
        # move dataset (without batching) to correct device
        dataset.to(device)
        # normalize dataset for pykan
        dataset.normalize_for_pykan()
        # update the model's grid from train samples
        model.update_grid_from_samples(dataset.data["train_input"])
        # pykan needs an eval-loss function that can be executed **without** any parameters. So the dataset (test split)
        # has to be hard-coded into the loss function for some reason...
        def hardcoded_pykan_testloss_metric():
            """Why would anybody program it like this?! Why not just make test GT and Prediction a parameter
            like every other loss function?..."""
            return loss_fn_backprop(model(dataset.data["test_input"].to(device)), dataset.data["test_label"].to(device))
        # fit the model
        model.fit(dataset=dataset.data, opt="LBFGS", steps=epochs, loss_fn=loss_fn_backprop, batch=batch_size,
                  metrics = [hardcoded_pykan_testloss_metric], display_metrics=["hardcoded_pykan_testloss_metric"])
        # evaluate the trained model on all loss functions
        train_losses, test_losses = eval_model(model, loss_fns, train_data=dataset.data["train_input"],
                                    test_data=dataset.data["test_input"],
                                    train_label=dataset.data["train_label"],
                                    test_label=dataset.data["test_label"],
                                    add_softmax_lastlayer=add_softmax_lastlayer)
        print(
            f"Final Train Loss: {[(lfn, l.item()) for lfn, l in train_losses.items()]}, Final Test Loss: {[(lfn, l.item()) for lfn, l in test_losses.items()]}")
        return train_losses, test_losses
    # if model is not PyKAN Wrapper, check if batch_size is > 0 (for pykan batch size should be -1)
    assert batch_size > 0
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=0)
    print("Number of trainable parameters in the model: ", get_num_parameters(model))

    optimizer = torch.optim.AdamW(model.parameters(), lr=0.01)
    # decay learning rate by 0.6 every epochs//10 steps
    scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=epochs // 10, gamma=0.6)

    # train loop
    for epoch in range(epochs):
        # evaluate without grads
        with torch.no_grad():
            if epoch % logging_interval == 0:
                train_losses, test_losses = eval_model(model, loss_fns, train_data=dataset.data["train_input"].to(device),
                                                       test_data=dataset.data["test_input"].to(device),
                                                       train_label=dataset.data["train_label"].to(device),
                                                       test_label=dataset.data["test_label"].to(device),
                                                       add_softmax_lastlayer=add_softmax_lastlayer)
                print(
                    f"Epoch {epoch} Train Loss: {[(lfn, l.item()) for lfn, l in train_losses.items()]}, Test Loss: {[(lfn, l.item()) for lfn, l in test_losses.items()]}, LR: {scheduler.get_last_lr()}")

        # iterate over all batches in train dataloader
        for batch, (X, y) in enumerate(dataloader):
            X = X.to(device)
            y = y.to(device)
            # potentially convert model's output to real
            if last_layer_output_real:
                train_predictions = model(X).real
            else:
                train_predictions = model(X)
            # calculate loss for backprop
            train_loss = loss_fn_backprop(train_predictions, y)
            if sparsify:  # potentially regularize the net using KANExplainer (Edge's relevance scores)
                sparsity_regularization = 0
                kan_explainer.calc_relevances_pykan()
                for k in kan_explainer.edge_relevances.keys():
                    sparsity_regularization += 1*(kan_explainer.get_edge_relevance(k).abs()).sum()
                train_loss = train_loss + sparsity_regularization
            train_loss.backward()
            optimizer.step()
            optimizer.zero_grad()
        scheduler.step()
    # Final evaluation
    train_losses, test_losses = eval_model(model, loss_fns, train_data=dataset.data["train_input"].to(device),
                                           test_data=dataset.data["test_input"].to(device),
                                           train_label=dataset.data["train_label"].to(device),
                                           test_label=dataset.data["test_label"].to(device),
                                           add_softmax_lastlayer=add_softmax_lastlayer)
    print(
        f"Final Train Loss: {[(lfn, l.item()) for lfn, l in train_losses.items()]}, Final Test Loss: {[(lfn, l.item()) for lfn, l in test_losses.items()]}")
    return train_losses, test_losses


"""
File: PyKANWrapper.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Wrapper for pyKAN to make it compatible with our KanPlotter and KanExplainer.
"""
from datetime import datetime
import numpy as np
import torch

from .WrapperTemplate import WrapperTemplate
from kan import KAN as PyKAN


class PyKANWrapper(PyKAN, WrapperTemplate):
    """
    Wrapper class for PyKAN model to work with our Plotting and Explainer Classes

    Parameter-Counts in PyKAN Layer (requires grad):
    in*out					# Weight for Basis-Scale
    in*out					# Weight for Spline-Scale
    in * out * (num_grids + k)		# Spline-Coefficients
    """

    def __init__(self, **kwargs):
        # Rename Keys to match with other KAN implementations
        # change 'layers_hidden' to 'width'
        kwargs["width"] = kwargs["layers_hidden"]
        self.layers_hidden = kwargs["layers_hidden"]
        del kwargs["layers_hidden"]
        # change 'num_grids' to 'grid'
        kwargs["grid"] = kwargs["num_grids"]
        self.num_grids = kwargs["num_grids"]
        del kwargs["num_grids"]

        # extract parameter 'update_grid' if exists and remove it from kwargs (kwargs will be later forwarded to KAN)
        if "update_grid" in kwargs:
            self._update_grid = kwargs["update_grid"]
            del kwargs["update_grid"]
        else:
            self._update_grid = False
        # set use_norm just for logging of experimental results
        self.use_norm = "grid_update" if self._update_grid else "no"

        # replace 'spline_order' with 'k'
        if "spline_order" in kwargs:
            kwargs["k"] = kwargs["spline_order"]
            del kwargs["spline_order"]
        # create random seed
        if "seed" not in kwargs:
            kwargs["seed"] = int(datetime.now().timestamp())
        # forward remaining kwargs to original KAN implementation
        super().__init__(**kwargs, auto_save=False)
        self.layer_widths = kwargs['width']
        # make sure it's a plain KAN without Multiplication Nodes
        for v in self.layer_widths:
            if v[1] != 0:  # v is a tuple (num_add_nodes, num_mult_nodes))
                raise NotImplementedError('PyKANWrapper: Multiplication Nodes in PyKAN are not yet supported')
        # dictionary for storing Edge and Node Activations
        self.edge_activations = dict()
        self.node_activations = dict()
        self.layers_hidden = [w[0] for w in self.layer_widths]
    def plot_curve(self, l_i_j_index: tuple, num_pts=1000) -> (np.ndarray, tuple[np.ndarray]):
        l,i,j = l_i_j_index
        # extract inputs and outputs from Node and Edge Activations
        inputs = self.get_node_activations((l,i)).cpu().detach().numpy()
        outputs = self.get_edge_activations((l,i,j)).cpu().detach().numpy()
        # they are not ordered yet; order them the same way (sorted based on 'inputs')
        rank = np.argsort(inputs)
        inputs = inputs[rank]
        outputs = outputs[rank]
        # ToDo: fill None, None with basis function and individual splines if required (only visualization)
        return inputs, (None, None, outputs)

    def get_edge_activations(self, l_i_j_index: tuple):
        layer, i, j = l_i_j_index
        return self.spline_postacts[layer][:,j,i]  # yes, j and i are in 'wrong' order, this is correct

    def get_node_activations(self, l_i_index: tuple):
        layer, i = l_i_index
        return self.acts[layer][:,i]


    def get_layersizes(self) -> list:
        return [w[0] for w in self.layer_widths]

    def get_gridrange(self) -> list:
        raise NotImplementedError("Grid Range return in pykan not yet supported")

    def get_kan_layers(self) -> torch.nn.ModuleList:
        return self.act_fun

    def gather_activations(self, x):
        self.forward(x)

    def fit(self, **kwargs):
        # insert 'update_grid' parameter from initialization time
        if "update_grid" not in kwargs:
            kwargs['update_grid'] = self._update_grid
        super().fit(**kwargs)



"""
File: plot_kan.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Plot KAN models (real-valued as well as complex-valued) with interactive elements.
Features:
- Transparency can be adjusted with a slider (with respect to the calculated relevance scores by KANExplainer)
- double-click on nodes only shows incoming and outgoing edge of that specific node. Can be undone by double-clicking
  anywhere else inside the plot
- the function of one single edge can be visualized as a close-up detailed version on the right. Selection is done by
  entering the tuple of the edge (l,i,j) for edge E_{l,i,j} in the text-box and pressing enter.
"""
import math
import matplotlib.colors
import numpy as np
import torch
from matplotlib import pyplot as plt
from matplotlib.widgets import Slider, CheckButtons, TextBox

from ...models.wrapper import WrapperTemplate
from ..explain_kan import KANExplainer, DummyKANExplainer
from . import cplotting_tools as cplot

# Default Values for Plot sizes
_NODE_PLOT_CIRCLESIZE = 0.05  # Node circle size
_ACTPLOT_SIZE = (0.7,0.7)  # size of plotted activation functions within network plot (rectangular sub-plot)
_PLOT_AT_LINK_FRACTION = 0.5  # Plot activation function at this fraction of a connection's link (0.5 = at middle of vertex)
_COLORS = list(matplotlib.colors.BASE_COLORS)  # colormap for nodes and their outgoing edges within each layer
_COLORS.remove("w")  # plotting in white is pointless
_LINESTYLES = ["solid", "dashed", "solid"]  # for base, splines, sum (real-valued only)

matplotlib.use('TkAgg')  # use a Backend that supports interactive elements

class KANPlot():
    def __init__(self, model: WrapperTemplate, kan_explainer: KANExplainer=None, input_featurenames=None, output_names=None, plot_options=None, complex_valued=False):
        """
        :param model: Model to Plot (should implement the methods of WrapperTemplate)
        :param kan_explainer: For relevance-scores per edge (set to None to disable KAN explainer)
        :param input_featurenames: Input Featurenames displayed in the plot
        :param output_names: Output Names displayed in the plot
        :param plot_options: Boolean Array for default visibility of components (just leave at None and manually play around with the Checkboxes in the plot)
        :param complex_valued: Switch to plotting a Complex-Valued KAN
        """
        self.model = model
        self.model.to(torch.device("cpu"))
        if kan_explainer is None:
            # if no explainer is given, use a dummy explainer that always outputs node's and edge's relevance as 1
            self.kan_explainer = DummyKANExplainer()
        else:
            self.kan_explainer = kan_explainer
        self.input_featurenames = input_featurenames
        self.output_names = output_names
        self.complex_valued = complex_valued
        # scale up x-axis to make the plot less elongated
        self.xscale_factor = len(self.model.get_kan_layers()) + 2
        # +0.5 for centering, * _XSCALE_FACTOR to make plot wider and square-ish (otherwise plot would be super elongated and hard to read)
        # using same axis scaling alone is not enough since that makes circles ellipses and stretches activation function plots
        self.nodeid2xpos = lambda k, dimsize: (k + 0.5) / dimsize * self.xscale_factor  # node id within layer to xpos
        self.xpos2nodeid = lambda x, dimsize: (x / self.xscale_factor * dimsize) - 0.5  # xpos to node id within one layer
        # for converting relevance score to transparency
        self.beta = 0.3
        # dictionary containing the plotted activation functions with key [function_type][(l,i,j)]
        self.act_functs = dict()
        # and their corresponding coordinate axes
        self.act_functs_coordinatesystems = dict()
        # links between nodes with keys (l,i,j)
        self.connections = dict()
        # textboxes for input featurenames with keys (i)
        self.textboxes_input_featurenames = dict()
        self.edge_relevances = dict()  # key (l,i,j) tuple
        self.node_relevances = dict()  # key: (l,i) tuple
        # nodes with key (l,i)
        self.nodes = dict()
        # Dict for storing images (screenshots) in complex valued mode or potentially lists of matplotlib Line objects
        self.complex_images = dict()
        self.complex_draw_colorbar = True  # 'counter' to plot colorbar only once
        self.plot_options = plot_options
        self.lij_last_choice = (0,0,0)  # function to plot in detail view on the right (remember choice for updates)
        if self.plot_options is None:
            # plot options exist for base, splines, act_funct and coordinate_axes
            # base is i.e. SiLu, splines are the individual functions around grid-points (i.e. splines or RBFs),
            # act_funct is the sum of everything and coordinate_axes are 2D coordinate axes to be plotted (only real-valued)
            self.plot_options = [False, False, True, True]
        self.function_types = ["base", "splines", "act_funct", "coordinate_axes"]
        for ft in self.function_types:
            self.act_functs[ft] = dict()
        # remember which visibilities have been updated to revert back to them
        self.visibilities_to_revert_to = []
        # which node is currently in focus (double-clicked)
        self.currently_focused_node = None

        if self.complex_valued:
            # if complex-valued the right detail plot needs to be a 3d projection
            fig = plt.figure(figsize=(24, 8))
            axs = [None, None]
            axs[0] = fig.add_subplot(1, 2, 1)
            axs[1] = fig.add_subplot(1, 2, 2, projection='3d')
        else:
            # for real-valued just create 2 subplots
            fig, axs = plt.subplots(figsize=(8, 8), ncols=2, nrows=1, width_ratios=[2,1])
        axs[0].axis("on")
        axs[0].set_aspect("equal")
        self.fig = fig
        self.axs = axs

        if not self.complex_valued:
            # Normalize the functions per Layer to make the plot more readable
            # it might also be better to turn this off sometimes (just append 1.0 instead of max(abs(act_ys))
            self.max_activation_value = dict()  # this is supposed to store the max activation value per layer
            for layer_id, layer in enumerate(model.get_kan_layers()):
                # create empty list for current layer
                self.max_activation_value[layer_id] = []
                for i in range(self.model.get_layersizes()[layer_id]):
                    for j in range(self.model.get_layersizes()[layer_id+1]):
                        act_xs, act_ys = self.model.plot_curve((layer_id,i,j))
                        act_ys = act_ys[2]  # only care about sum of splines and silu (not the individual components themselves)
                        # append max of each edge's activation function within this layer
                        self.max_activation_value[layer_id].append(max(abs(act_ys)))
            for k in self.max_activation_value.keys():  # take max within each layer k
                self.max_activation_value[k] = np.max(self.max_activation_value[k])
        # Create Interactive Sliders and Checkboxes
        slider_ax = plt.axes([0.05, 0.2, 0.05, 0.6])
        # Sliders and Checkboxes ***MUST*** be stored as attributes of the class
        # otherwise stuff breaks (maybe the Garbage Collector deletes them, although they should be used later on?)...
        # took me ~6 hours to find this out. Amazing.
        self.slider = Slider(slider_ax, label="beta", valmin=0, valmax=2, valinit=0.5, orientation="vertical")
        # if slider value changes, call self.update_transparency() with new slider value as parameter
        self.slider.on_changed(self.update_transparency)
        # checkboxes for switching on/off the individual function types. Only implemented for real-valued functions
        if not self.complex_valued:
            cb_ax = plt.axes([0.2, 0.8, 0.2, 0.2])
            self.cb_buttons = CheckButtons(ax=cb_ax, labels=self.function_types, actives=self.plot_options)
            self.cb_buttons.on_clicked(self.update_plot_options)
    def plot_all(self):
        """
        Plot the entire model together with detailed view on the right
        """
        # plot model structure
        self.plot_model()
        # and one single activation function in detail
        self.plot_single_function()
        plt.show()
        #plt.savefig("figure.svg", format="svg", dpi=600)

    def calc_opaqueness(self, relevance):
        """
        Calculate the opaqueness of an Edge / Node based on it's relevance score
        """
        result = self.beta * math.tanh(relevance)  # same as for pyKAN 2.0
        #result = self.beta * math.tanh(relevance/50)  # scale by factor 50 to make smaller differences more visible (good for the knot dataset plots)
        return np.clip(result, 0, 1)  # clip to range [0,1]
    def update_transparency(self, slider_value=None):
        """
        Updates the transparency of Edges / Nodes, whenever the Slider is changed
        """
        self.undo_visibility_focus()
        if slider_value is not None:
            self.beta = slider_value
        # iterate through all the visible components in the plot and change their alpha values
        # activation functions (each function_type)
        for function_type in self.act_functs.keys():
            for l_i_j_index in self.connections.keys():
                if function_type in self.act_functs and l_i_j_index in self.act_functs[function_type]:
                    for spline_function in self.act_functs[function_type][l_i_j_index]:
                        spline_function.set_alpha(self.calc_opaqueness(self.kan_explainer.get_edge_relevance(l_i_j_index)))
                        # update their visibilities as desired (invisible / visible) as self.plot_options dictates
                        if not self.plot_options[self.function_types.index(function_type)]:
                            spline_function.set_visible(False)
                        else:
                            spline_function.set_visible(True)
                # set alpha of vertices
                self.connections[l_i_j_index].set_alpha(self.calc_opaqueness(self.kan_explainer.get_edge_relevance(l_i_j_index)))
        # also update the plotted coordinatesystems
        for k in self.act_functs_coordinatesystems.keys():
            if not self.plot_options[3]:
                for gridline in self.act_functs_coordinatesystems[k]:
                    gridline.set_visible(False)
            else:
                for gridline in self.act_functs_coordinatesystems[k]:
                    gridline.set_visible(True)
                    gridline.set_alpha(self.calc_opaqueness(self.kan_explainer.get_edge_relevance(k)))
        # update nodes
        for k in self.nodes.keys():
            self.nodes[k].set_alpha(self.calc_opaqueness(self.kan_explainer.get_node_relevance(k)))
        # update textboxes (disabled for now)
        for k in self.textboxes_input_featurenames.keys():
            pass
            #self.textboxes_input_featurenames[k].set_alpha(self.calc_opaqueness(self.kan_explainer.get_node_relevance((0,k))))
        # update screenshots of complex-valued functions (or the lineset objects)
        for k in self.complex_images.keys():
            if type(self.complex_images[k]) == list:  # list of lineset objects
                for el in self.complex_images[k]:
                    el.set_alpha(self.calc_opaqueness(self.kan_explainer.get_edge_relevance(k)))
            else:  # images (single item)
                self.complex_images[k].set_alpha(self.calc_opaqueness(self.kan_explainer.get_edge_relevance(k)))
            self.connections[k].set_alpha(self.calc_opaqueness(self.kan_explainer.get_edge_relevance(k)))
        self.fig.canvas.draw_idle()
        self.focus_on_node()  # keep focused node still focused
    def update_plot_options(self, label):
        """This method is called whenever the Checkboxes are touched. Update the visibility of the Activation Function's
        components according to the selection
        :param label: new plot options (id of the checkbox changed)
        """
        # flip plot_option for this function_type
        self.plot_options[self.function_types.index(label)] = not self.plot_options[self.function_types.index(label)]
        # update the transparencies and visibilities
        self.update_transparency()
        # also update the single Activation Function plot (detail view) based on the Checkbox selection
        self.update_single_function(l_i_j_index=self.lij_last_choice)
    def plot_model(self):
        """
        Plot the Model structure (left)
        """
        # plot all layers
        for i in range(len(self.model.get_kan_layers())):
            current_layer = self.model.get_kan_layers()[i]
            self.plot_layer(layer_id=i)

        # plot feature names
        if self.input_featurenames is not None:
            # add linebreaks in very long featurenames...
            input_featurenames = [x.split(".")[-1] for x in self.input_featurenames]
            for i in range(len(input_featurenames)):
                linebroken = ""
                for j, c in enumerate(input_featurenames[i]):
                    linebroken += c
                    if j % 18 == 0 and j > 0:
                        linebroken += "\n"
                input_featurenames[i] = linebroken
            assert self.model.get_layersizes()[0] == len(input_featurenames)
            # plot input featurenames
            for i in range(self.model.get_layersizes()[0]):
                textbox_xpos = self.nodeid2xpos(i, self.model.get_layersizes()[0])
                self.textboxes_input_featurenames[i] = self.axs[0].text(textbox_xpos, -0.5, input_featurenames[i], horizontalalignment='center',
                         verticalalignment='center', rotation=90, fontsize=12, bbox=dict(facecolor='none', edgecolor='red'))
        # plot output names
        for i in range(self.model.get_layersizes()[-1]):
            if self.output_names is not None:
                assert self.model.get_layersizes()[-1] == len(self.output_names)
                textbox_xpos = self.nodeid2xpos(i, self.model.get_layersizes()[-1])
                self.axs[0].text(textbox_xpos, len(self.model.get_kan_layers()) + 0.25, self.output_names[i], horizontalalignment='center',
                                 verticalalignment='center')
            # plot output node(s)
            output_node_circle = plt.Circle((self.nodeid2xpos(i, self.model.get_layersizes()[-1]), len(self.model.get_layersizes()) - 1),
                                            radius=_NODE_PLOT_CIRCLESIZE,
                                            color="black")
            self.nodes[len(self.model.get_layersizes()) - 1, i] = output_node_circle
            self.axs[0].add_patch(output_node_circle)

        # set ylim to not cut away feature names or output names
        self.axs[0].set_ylim(-1, len(self.model.get_kan_layers()) + 0.5)
        # increase xlim slightly
        current_xlim = self.axs[0].get_xlim()
        self.axs[0].set_xlim(current_xlim[0]-0.2, current_xlim[1]+0.2)
        # clicks (later checked if it was double-click) lead to focus on single node
        self.fig.canvas.mpl_connect('button_press_event', self.on_mouse_press)
        self.update_transparency()
    def focus_on_node(self):
        """
        Only display incoming and outgoing edges of self.currently_focused_node
        (which is set at another place in the code).
        """
        # revert changes for focusing on previous node (might already be focusing on some node)
        self.undo_visibility_focus()
        if self.currently_focused_node is None:  # no new node to focus on
            return
        # node to focus on
        layer_id, node_id = self.currently_focused_node
        # Turn every edge that's not incoming or outgoing of Node (layer_id, node_id) invisible
        for (l, i), node in self.nodes.items():
            if i == node_id and l == layer_id:  # node within current layer should stay visible
                continue
            if layer_id == l - 1 or layer_id == l + 1:  # Nodes in the previous and next Layer should also stay visible
                continue
            # remember the changes done so we can go back to the original plot later on
            self.visibilities_to_revert_to.append((node, node.get_visible()))
            node.set_visible(False)
        # do the same for activation functions
        for ft in self.function_types:
            for (l, i, j), act_funs in self.act_functs[ft].items():
                if (l == layer_id and i == node_id) or (l == layer_id - 1 and j == node_id):
                    continue
                for act_fun in act_funs:
                    self.visibilities_to_revert_to.append((act_fun, act_fun.get_visible()))
                    act_fun.set_visible(False)
        # and edges
        for (l, i, j), edge in self.connections.items():
            if (l == layer_id and i == node_id) or (l == layer_id - 1 and j == node_id):
                continue
            self.visibilities_to_revert_to.append((edge, edge.get_visible()))
            edge.set_visible(False)
            if (l,i,j) in self.complex_images:
                lines = self.complex_images[(l,i,j)]
                if type(lines) != list:
                    lines = [lines]
                for line in lines:
                    self.visibilities_to_revert_to.append((line, line.get_visible()))
                    line.set_visible(False)
        # and the coordinate systems plotted inside the model's structure
        for (l, i, j), act_fun_coord_systems in self.act_functs_coordinatesystems.items():
            if (l == layer_id and i == node_id) or (l == layer_id - 1 and j == node_id):
                continue
            for act_fun_coord_sys in act_fun_coord_systems:
                self.visibilities_to_revert_to.append((act_fun_coord_sys, act_fun_coord_sys.get_visible()))
                act_fun_coord_sys.set_visible(False)

        self.fig.canvas.draw_idle()
    def on_mouse_press(self, event):
        """This function is called whenever a mouse-click happens. It tries to match the double click's position to a node
        and then makes every Edge / Node, that is not directly connected to it, invisible.
        :param event: mouse click event
        """
        if not event.dblclick:  # Has to be double click
            return
        # wherever the click happens: undo the last focus on node
        self.undo_visibility_focus()
        if event.xdata is None or event.ydata is None:
            return
        self.currently_focused_node = None
        # Try to convert event.ydata to layer index
        # double-click has to happen within node circle vertically (so we first match the layer)
        if abs(round(event.ydata) - event.ydata) <= _NODE_PLOT_CIRCLESIZE:
            layer_id = round(event.ydata)
        else:
            print("failed to match layer")
            return
        # Try to find matching node within this layer
        node_candidate_id = self.xpos2nodeid(event.xdata, self.model.get_layersizes()[layer_id])
        node_candidate_id = int(node_candidate_id)
        node_candidate_xpos = self.nodeid2xpos(node_candidate_id, self.model.get_layersizes()[layer_id])
        # if nearest node is close enough to mouse pointer (horizontally)
        if abs(node_candidate_xpos - event.xdata) <= _NODE_PLOT_CIRCLESIZE:
            node_id = round(node_candidate_id)
        else:
            print("failed to match node")
            return
        # set node to focus on
        self.currently_focused_node = (layer_id, node_id)
        self.focus_on_node()  # actually do the visibility updates
    def undo_visibility_focus(self):
        # Undo all the visibility changes made by on_mouse_press
        for line, state in self.visibilities_to_revert_to:
            line.set_visible(state)
        self.visibilities_to_revert_to = []
        self.fig.canvas.draw_idle()
    def plot_act_funct(self, x_center,y_center,data_x,functions_y, l_i_j_index, color="black"):
        """
        Plot a single Activation Function inside the Model's structure plot
        :param x_center: plot around xcenter
        :param y_center: plot around ycenter
        :param data_x: x-positions of the activation function
        :param functions_y: y-values of the activation function as an array [base-function, splines, resulting act function]
        :param l_i_j_index: layer,i,j index of the corresponding Edge
        :param color: color to plot the function in
        """
        convert_x_to_xpos_global = lambda x: _ACTPLOT_SIZE[0] * x + x_center
        convert_y_to_ypos_global = lambda y: y / torch.amax(torch.abs(y)) * _ACTPLOT_SIZE[1] + y_center
        # convert_y_to_ypos_global = lambda y: y / self.max_activation_value[l_i_j_index[0]] * _ACTPLOT_SIZE[1] + y_center

        if self.complex_valued:  # complex-valued functions are treated differently
            rect = [x_center-_ACTPLOT_SIZE[0], y_center-_ACTPLOT_SIZE[0],_ACTPLOT_SIZE[0], _ACTPLOT_SIZE[1]]
            rect = [x_center - _ACTPLOT_SIZE[0], x_center + _ACTPLOT_SIZE[0], y_center - _ACTPLOT_SIZE[1], y_center + _ACTPLOT_SIZE[1]]
            # get the function's values
            (xs, ys), (basis, rbfs, zs) = self.model.plot_curve(l_i_j_index, num_pts=100)
            # plot the complex-valued function and return as image / screenshot
            img = cplot.complex_plot3D(xs, ys, zs, draw_colorbar=False, return_image=True, fontsize=32)
            # embed the screenshot into the current plot
            displayed_img = self.axs[0].imshow(img, extent=rect)
            # and store it in self.complex_images for later transparency updates
            self.complex_images[l_i_j_index] = displayed_img
        else:  # real-valued
            self.act_functs_coordinatesystems[l_i_j_index] = []
            for ft in self.function_types:
                self.act_functs[ft][l_i_j_index] = []
            # create a coordinate system (x- and y-axis) for this specific activation function
            yaxis = self.axs[0].plot([x_center, x_center], [-_ACTPLOT_SIZE[1] + y_center, _ACTPLOT_SIZE[1] + y_center],
                                          color="black", linewidth=0.5)[0]
            xaxis_left = convert_x_to_xpos_global(-0.5)
            xaxis_right = convert_x_to_xpos_global(0.5)
            xaxis = self.axs[0].plot([xaxis_left, xaxis_right], [y_center, y_center],
                                          color="black", linewidth=0.5)[0]
            self.act_functs_coordinatesystems[l_i_j_index].append(yaxis)
            self.act_functs_coordinatesystems[l_i_j_index].append(xaxis)

            # plot all available activation functions
            # this can be [base-function (i.e. SiLU), splines / learnable functions, sum of all of them]
            for data_y_index, data_y in enumerate(functions_y):
                if data_y is None:
                    continue
                # make data fit into the plotted coordinate system (scale and shift)
                data_x_shifted = convert_x_to_xpos_global(data_x)
                data_y_shifted = convert_y_to_ypos_global(data_y)
                # one of [base-function, learnable function, resulting act function]
                curr_func_type = self.function_types[data_y_index]
                # if learnable function, data_y is an array containing all the individual learned functions
                # (centered around different grid-points)
                if data_y_index == 1:
                    for i in range(data_y_shifted.shape[1]):
                        self.act_functs[curr_func_type][l_i_j_index].append(
                            self.axs[0].plot(data_x_shifted, data_y_shifted[:, i], color=color,
                                                  linewidth=0.5 * (data_y_index + 1), linestyle=_LINESTYLES[data_y_index])[0])
                else:  # otherwise just plot single function

                    self.act_functs[curr_func_type][l_i_j_index].append(
                        self.axs[0].plot(data_x_shifted, data_y_shifted, color=color,
                                              linewidth=0.5 * (data_y_index + 1), linestyle=_LINESTYLES[data_y_index])[0])



    def plot_layer(self, layer_id):
        """
        Plot one layer of the model
        :param layer_id: ID of the layer
        """
        # incoming and outgoing number of nodes
        in_d = self.model.get_layersizes()[layer_id]
        out_d = self.model.get_layersizes()[layer_id+1]
        # plot every Edge between this one and the next layer
        for i in range(0, in_d):
            # calculate node coordinates
            node_i_xpos = self.nodeid2xpos(i, in_d)
            for j in range(0, out_d):
                # calculate node coordinates
                node_j_xpos = self.nodeid2xpos(j, out_d)
                # get activation function values from the model
                act_xs, act_ys = self.model.plot_curve((layer_id,i,j))
                # calculate position for plotting the activation function at
                link_length = math.sqrt(1 + (node_i_xpos - node_j_xpos)**2)  # 1² + (x_i - x_j)² = hypothenuse
                plot_at_link_length = _PLOT_AT_LINK_FRACTION*link_length
                link_angle = math.asin(1/link_length)
                plot_at_ypos = math.sin(link_angle) * plot_at_link_length + layer_id
                factor = -1 if node_j_xpos < node_i_xpos else 1
                # add very small offset 0.02 to have nicer plot
                plot_at_xpos = factor * math.cos(link_angle) * plot_at_link_length + node_i_xpos - 0.02
                # plot activation function using new axis
                self.plot_act_funct(plot_at_xpos, plot_at_ypos, act_xs, act_ys, (layer_id, i, j), color=_COLORS[i%len(_COLORS)])
                # plot connecting line between the nodes in subsequent functions
                self.connections[(layer_id,i,j)] = self.axs[0].plot([node_i_xpos, node_j_xpos],
                                                                         [layer_id + _NODE_PLOT_CIRCLESIZE, layer_id + 1 - _NODE_PLOT_CIRCLESIZE], color="black",
                                                                         linewidth=1)[0]

            # Draw circle for nodes in current input layer
            node_i_circle = plt.Circle((node_i_xpos, layer_id), radius=_NODE_PLOT_CIRCLESIZE, color=_COLORS[i%len(_COLORS)])
            self.nodes[(layer_id,i)] = node_i_circle
            self.axs[0].add_patch(node_i_circle)
    def plot_single_function(self):
        """
        Plot one single activation function in detail (right window part)
        """
        lij_chooser_ax = self.fig.add_axes([0.55, 0.9, 0.1, 0.1])
        # text-box to select which function should be plotted in detail
        self.lij_chooser = TextBox(lij_chooser_ax, 'l,i,j', initial="0,0,0")
        self.lij_chooser.on_submit(self.update_single_function)
        self.update_single_function(self.lij_chooser.text)
    def update_single_function(self, l_i_j_index):
        """
        Update the window with the one single activation function in detail.
        :param l_i_j_index: layer, i and j for the selected Activation function to plot as tuple (layer, i, j)
        """
        l, i, j = [int(idx) for idx in l_i_j_index.split(",")]
        self.lij_last_choice = l_i_j_index  # remember choice
        self.axs[1].cla()  # clear axis

        if self.complex_valued:  # complex-valued
            (xs, ys), (basis, rbfs, zs) = self.model.plot_curve((l,i,j))
            # plot directly into the axis of the right subfigure
            cplot.complex_plot3D(xs, ys, zs, fig_ax=(self.fig, self.axs[1]), draw_colorbar=self.complex_draw_colorbar)
            # draw the colorbar only once (otherwise they stack over each other weirdly...)
            if self.complex_draw_colorbar:
                self.complex_draw_colorbar = False

        else:  # real-valued
            # colors for base, learnable functions, resulting act function
            colors = ["black","green","blue"]
            try:
                xs, ys = self.model.plot_curve((l,i,j))
                # plot options has boolean for every function_type ["base", "splines", "act_funct", "coordinate_axes"]
                for i in range(len(self.plot_options)-1):
                    if self.plot_options[i]:  # if current function_type should be plotted
                        self.axs[1].plot(xs, ys[i], linestyle=_LINESTYLES[i], linewidth=i+1, color=colors[i])
                # create coordinate axis at x=0 and y=0
                self.axs[1].axhline(y=0, color='k', linewidth=0.5)
                self.axs[1].axvline(x=0, color='k', linewidth=0.5)
                self.fig.canvas.draw_idle()
            except ValueError:
                print("Wrong choice for l,i,j input! (needs to be 3 integers separated by comma). Given:", l_i_j_index)
            except IndexError:
                print("Wrong choice for l,i,j input! (Edge l,i,j needs to exist). Given:", l_i_j_index)

